#!/usr/bin/env python
# -*- coding: utf-8 -*-


from covsirphy.ode.mbase import ModelBase
from covsirphy.phase.phase_unit import PhaseUnit
import warnings
import pytest
from covsirphy import SIRF, PhaseSeries, ParamTracker


@pytest.fixture(scope="module")
def tracker(jhu_data, population_data):
    population = population_data.value(country="Japan")
    record_df = jhu_data.subset(country="Japan", population=population)
    series = ParamTracker.create_series("01Apr2020", "01Nov2020", population)
    assert isinstance(series, PhaseSeries)
    return ParamTracker(
        record_df=record_df, phase_series=series, area="Japan"
    )


class TestParamTracker(object):
    def test_series(self, tracker):
        assert isinstance(tracker.series, PhaseSeries)
        assert len(tracker) == len(tracker.series)

    def test_add(self, tracker):
        tracker.add(days=10)
        tracker.add(end_date="01Jun2020")
        tracker.add(days=100)
        tracker.add()
        with pytest.raises(ValueError):
            tracker.add(end_date="01May2020")

    def test_delete(self, tracker):
        length = len(tracker)
        tracker.delete(phases=["1st", "last"])
        assert len(tracker) == length - 2

    def test_delete_all(self, tracker):
        assert tracker
        tracker.delete_all()
        assert not tracker

    def test_combine(self, tracker):
        tracker.add("01May2020")
        tracker.add("01Jun2020")
        tracker.add("01Jul2020")
        tracker.add("01Aug2020")
        tracker.combine(phases=["2nd", "last"])
        tracker.combine(phases=["0th", "1st"])

    def test_before_trend(self, tracker):
        tracker.delete_all()
        with pytest.raises(ValueError):
            tracker.find_phase("01May2020")

    def test_trend(self, tracker):
        warnings.simplefilter("ignore", category=UserWarning)
        tracker.trend(show_figure=True)
        assert isinstance(tracker.change_dates(), list)

    def test_trend_find_phase(self, tracker):
        with pytest.raises(IndexError):
            tracker.find_phase("01Jan2000")
        tracker.find_phase("01May2020")

    @pytest.mark.parametrize(
        "date", ["01Apr2020", "01May2020", "01Jun2020", "01Jul2020"])
    def test_separate(self, tracker, date):
        close_dates = tracker.near_change_dates()
        if date in close_dates:
            with pytest.raises(ValueError):
                tracker.separate(date)
        else:
            tracker.separate(date)

    def test_past_phases(self, tracker):
        all_phases, all_units = tracker.past_phases()
        sel_phases, sel_units = tracker.past_phases(
            phases=["0th", "1st"])
        assert set(sel_phases) == set(["0th", "1st"])
        assert set(sel_phases).issubset(all_phases)
        assert set(sel_units).issubset(all_units)

    def test_future_phases(self, tracker):
        tracker.trend()
        phases0, units0 = tracker.future_phases()
        assert len(phases0) == len(units0) == 0
        tracker.add(end_date="01Apr2021")
        phases, units = tracker.future_phases()
        assert len(phases) == len(units) == 1
        assert isinstance(units[0], PhaseUnit)
        tracker.delete(phases=phases)

    def test_before_estimate(self, tracker):
        with pytest.raises(NameError):
            tracker.simulate()

    def test_estimate(self, tracker):
        tracker.estimate(
            SIRF, phases=["1st"], timeout=5, timeout_iteration=5)
        tracker.estimate(SIRF, timeout=5, timeout_iteration=5)
        with pytest.raises(IndexError):
            tracker.estimate(SIRF, timeout=5, timeout_iteration=5)

    def test_last_model(self, tracker):
        assert issubclass(tracker.last_model, ModelBase)

    def test_simulate(self, tracker):
        tracker.simulate()

    def test_score(self, tracker):
        # Scores of all phases
        metrics_list = ["MAE", "MSE", "MSLE", "RMSE", "RMSLE"]
        for metrics in metrics_list:
            score = tracker.score(metrics=metrics)
            assert isinstance(score, float)
        # Scores of target phases
        tracker.score(metrics="RMSLE", phases=["0th", "2nd"])
        # Errors with arguments
        with pytest.raises(TypeError):
            tracker.score(variables="Infected")
        with pytest.raises(KeyError):
            tracker.score(variables=["Susceptible"])
        with pytest.raises(ValueError):
            tracker.score(metrics="Subjective evaluation")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import functools
from multiprocessing import cpu_count, Pool
from covsirphy.util.stopwatch import StopWatch
from covsirphy.cleaning.term import Term
from covsirphy.cleaning.jhu_data import JHUData
from covsirphy.cleaning.population import PopulationData
from covsirphy.ode.mbase import ModelBase
from covsirphy.phase.phase_unit import PhaseUnit


class MPEstimator(Term):
    """
    Perform multiprocessing of Phaseunit.estimate()

    Args:
        model (covsirphy.ModelBase or None): ODE model
        jhu_data (covsirphy.JHUData): object of records
        population_data (covsirphy.PopulationData): PopulationData object
        record_df (pandas.DataFrame)
            Index:
                reset index
            Columns:
                - Date (pd.TimeStamp): Observation date
                - Confirmed (int): the number of confirmed cases
                - Infected (int): the number of currently infected cases
                - Fatal (int): the number of fatal cases
                - Recovered (int): the number of recovered cases
                - any other columns will be ignored
        tau (int or None): tau value [min], a divisor of 1440
        kwargs: keyword arguments of model parameters

    Notes:
        When @record_df is None, @jhu_data and @population_data must be specified.
    """

    def __init__(self, model, jhu_data=None, population_data=None,
                 record_df=None, tau=None, **kwargs):
        # Records
        if jhu_data is not None and population_data is not None:
            self.jhu_data = self.ensure_instance(
                jhu_data, JHUData, name="jhu_data")
            # Population
            self.population_data = self.ensure_instance(
                population_data, PopulationData, name="population_data")
            self.from_dataset = True
        else:
            self.record_df = self.ensure_dataframe(
                record_df, name="record_df", columns=self.NLOC_COLUMNS)
            self.from_dataset = False
        # Arguments
        self.model = self.ensure_subclass(model, ModelBase, "model")
        self._tau = self.ensure_tau(tau)
        self.param_dict = {
            k: v for (k, v) in kwargs.items() if k in model.PARAMETERS}
        self._units = []

    @property
    def tau(self):
        """
        int or None: tau value [min]
        """
        return self._tau

    def add(self, units):
        """
        Register PhaseUnits.

        Args:
            units (list[covsirphy.PhaseUnit]): list of phases

        Returns:
            covsirphy.MPEstimator: self
        """
        if not isinstance(units, list):
            raise TypeError("@units must be a list of PhaseUnits.")
        type_ok = all(isinstance(unit, PhaseUnit) for unit in units)
        if not type_ok:
            raise TypeError("@units must be a list of PhaseUnits.")
        units = [
            unit.set_ode(model=self.model, **self.param_dict) for unit in units
        ]
        self._units.extend(units)
        return self

    def _run(self, unit, tau, auto_complement=False, **kwargs):
        """
        Run estimation for one phase.

        Args:
            unit (covsirphy.PhaseUnit): unit of one phase
            tau (int or None): tau value [min], a divisor of 1440
            auto_complement (bool): if True and necessary, the number of cases will be complemented
            kwargs: keyword arguments of model parameters and covsirphy.Estimator.run()
        """
        # Set tau
        unit.set_ode(tau=tau)
        # Parameter estimation
        if self.from_dataset:
            id_dict = unit.id_dict.copy()
            try:
                country = id_dict["country"]
            except KeyError:
                raise KeyError(
                    "PhaseUnit.id_dict['country'] must have country name.")
            province = id_dict["province"] if "province" in id_dict else None
            population = self.population_data.value(
                country=country, province=province)
            record_df, _ = self.jhu_data.records(
                country=country, province=province, population=population,
                auto_complement=auto_complement)
        else:
            record_df = self.record_df.copy()
        unit.estimate(record_df=record_df, **kwargs)
        # Show the number of trials and runtime
        unit_dict = unit.to_dict()
        trials, runtime = unit_dict[self.TRIALS], unit_dict[self.RUNTIME]
        print(f"\t{unit}: finished {trials:>4} trials in {runtime}")
        return unit

    def run(self, n_jobs=-1, **kwargs):
        """
        Run estimation.

        Args:
            n_jobs (int): the number of parallel jobs or -1 (CPU count)
            kwargs: keyword arguments of model parameters and covsirphy.Estimator.run()

        Returns:
            list[covsirphy.PhaseUnit]
        """
        units = self._units[:]
        results = []
        # The number of parallel jobs
        n_jobs = cpu_count() if n_jobs == -1 else n_jobs
        # Start optimization
        print(f"\n<{self.model.NAME} model: parameter estimation>")
        print(f"Running optimization with {n_jobs} CPUs...")
        stopwatch = StopWatch()
        # Estimation of the last phase will be done to determine tau value
        if self._tau is None:
            unit_sel, units = units[-1], units[:-1]
            unit_est = self._run(unit=unit_sel, tau=None, **kwargs)
            self._tau = unit_est.tau
            results = [unit_est]
        # Estimation of each phase
        est_f = functools.partial(self._run, tau=self._tau, **kwargs)
        if n_jobs == 1:
            results = [est_f(unit) for unit in units]
        else:
            with Pool(n_jobs) as p:
                units_est = p.map(est_f, units)
            results.extend(units_est)
        # Completion
        stopwatch.stop()
        print(f"Completed optimization. Total: {stopwatch.stop_show()}")
        return results



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from covsirphy.cleaning.japan_data import JapanData
from covsirphy.util.error import SubsetNotFoundError
import pytest
import warnings
import pandas as pd
from covsirphy import CleaningBase, SIRF
from covsirphy import LinelistData, ExampleData, VaccineData
from covsirphy import Term, CountryData, PopulationPyramidData
from covsirphy import Word, Population


class TestLinelistData(object):

    def test_raw(self, linelist_data):
        assert isinstance(linelist_data.raw, pd.DataFrame)

    def test_linelist(self, linelist_data):
        with pytest.raises(NotImplementedError):
            linelist_data.total()
        assert isinstance(linelist_data.cleaned(), pd.DataFrame)
        assert isinstance(linelist_data.citation, str)

    @pytest.mark.parametrize("country", ["Japan", "Germany"])
    @pytest.mark.parametrize("province", [None, "Tokyo"])
    def test_subset(self, linelist_data, country, province):
        if (country, province) == ("Germany", "Tokyo"):
            with pytest.raises(KeyError):
                linelist_data.subset(country=country, province=province)
        else:
            df = linelist_data.subset(country=country, province=province)
            column_set = set(df) | set([Term.COUNTRY, Term.PROVINCE])
            assert column_set == set(LinelistData.LINELIST_COLS)

    @pytest.mark.parametrize("outcome", ["Recovered", "Fatal", "Confirmed"])
    def test_closed(self, linelist_data, outcome):
        if outcome in ["Recovered", "Fatal"]:
            linelist_data.closed(outcome=outcome)
        else:
            with pytest.raises(KeyError):
                linelist_data.closed(outcome=outcome)

    def test_recovery_period(self, linelist_data):
        assert isinstance(linelist_data.recovery_period(), int)


class TestObsoleted(object):
    def test_obsoleted(self):
        warnings.simplefilter("ignore", category=DeprecationWarning)
        Population(filename=None)
        Word()


class TestCleaningBase(object):
    def test_cbase(self):
        cbase = CleaningBase(filename=None)
        with pytest.raises(KeyError):
            cbase.iso3_to_country("JPN")
        with pytest.raises(NotImplementedError):
            cbase.total()
        cbase.citation = "citation"
        assert cbase.citation == "citation"


class TestExampleData(object):
    def test_iso3(self):
        example_data = ExampleData()
        example_data.add(SIRF, country="Japan")
        assert example_data.country_to_iso3("Japan") == "JPN"
        example_data.add(SIRF, country="Moon")
        assert example_data.country_to_iso3("Moon") == "---"

    def test_subset(self):
        example_data = ExampleData()
        example_data.add(SIRF, country="Japan")
        example_data.subset(country="Japan")
        example_data.subset_complement(country="Japan")
        example_data.records(country="Japan")


class TestCountryData(object):
    def test_cleaning(self, japan_data):
        assert isinstance(japan_data.raw_columns(), list)
        with pytest.raises(NotImplementedError):
            japan_data.set_variables()
        df = japan_data.cleaned()
        assert set(Term.COLUMNS).issubset(df.columns)

    def test_total(self, japan_data):
        df = japan_data.total()
        column_set = set(Term.VALUE_COLUMNS) | set(Term.RATE_COLUMNS)
        assert set(df.columns) == column_set

    def test_countries(self, japan_data):
        assert [japan_data.country] == japan_data.countries()

    def test_create(self, japan_data):
        country_data = CountryData(filename=None, country="Moon")
        with pytest.raises(ValueError):
            country_data.cleaned()
        country_data.raw = japan_data.raw
        country_data.set_variables(
            date="Date",
            confirmed="Positive",
            fatal="Fatal",
            recovered="Discharged",
            province="Area")
        df = country_data.cleaned()
        assert set(df.columns) == set(Term.COLUMNS)

    def test_create_province(self, japan_data):
        country_data = CountryData(
            filename=None, country="Moon", province="Reiner Gamma")
        country_data.raw = japan_data.raw
        country_data.set_variables(
            date="Date",
            confirmed="Positive",
            fatal="Fatal",
            recovered="Discharged",
            province=None)
        df = country_data.cleaned()
        assert set(df.columns) == set(Term.COLUMNS)

    def test_japan_meta(self, japan_data):
        raw_df = japan_data.meta(cleaned=False)
        assert isinstance(raw_df, pd.DataFrame)
        df = japan_data.meta(cleaned=True)
        assert set(df.columns) == set(JapanData.JAPAN_META_COLS)


class TestVaccineData(object):
    def test_cleaning(self, vaccine_data):
        df = vaccine_data.cleaned()
        assert set(VaccineData.VAC_COLS).issubset(df.columns)

    @pytest.mark.parametrize("country", ["Canada"])
    def test_subset(self, vaccine_data, country):
        df = vaccine_data.subset(country=country)
        assert set(df.columns) == set([Term.DATE, Term.VAC])
        clean_df = vaccine_data.cleaned()
        df = clean_df.loc[clean_df[Term.COUNTRY] == country, :]
        product = df.loc[df.index[0], Term.PRODUCT]
        vaccine_data.subset(
            country=country, product=product, start_date="15Dec2020", end_date="18Dec2020")
        with pytest.raises(SubsetNotFoundError):
            vaccine_data.subset(country=country, end_date="01May2020")

    @pytest.mark.parametrize("country", ["GBR"])
    def test_records(self, vaccine_data, country):
        vaccine_data.records(country=country)

    def test_total(self, vaccine_data):
        df = vaccine_data.total()
        assert set(df.columns) == set([Term.DATE, Term.VAC])


class TestPopulationPyramidData(object):
    def test_retrieve(self, pyramid_data):
        df = pyramid_data.retrieve("Japan")
        assert set(df.columns) == set(PopulationPyramidData.PYRAMID_COLS)
        with pytest.raises(SubsetNotFoundError):
            pyramid_data.retrieve("Moon")

    def test_cleaned(self, pyramid_data):
        df = pyramid_data.cleaned()
        assert set(df.columns) == set(PopulationPyramidData.PYRAMID_COLS)

    @pytest.mark.parametrize("country", ["Japan"])
    @pytest.mark.parametrize("sex", [None, "Female", "Male"])
    def test_records(self, pyramid_data, country, sex):
        df = pyramid_data.records(country, sex=sex)
        assert set(df.columns) == set(PopulationPyramidData.SUBSET_COLS)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import timedelta
import warnings
import numpy as np
import pandas as pd
import ruptures as rpt
from covsirphy.cleaning.term import Term
from covsirphy.phase.trend import Trend


class ChangeFinder(Term):
    """
    Find change points of S-R trend.

    Args:
        sr_df (pandas.DataFrame)
            Index:
                Date (pd.TimeStamp): Observation date
            Columns:
                - Recovered (int): the number of recovered cases (> 0)
                - Susceptible (int): the number of susceptible cases
                - any other columns will be ignored
        min_size (int): minimum value of phase length [days], over 2
        max_rmsle (float): minmum value of RMSLE score

    Notes:
        When RMSLE score > max_rmsle, predicted values will be None
    """

    def __init__(self, sr_df, min_size=5, max_rmsle=20.0):
        # Dataset
        self.sr_df = self.ensure_dataframe(
            sr_df, name="sr_df", time_index=True, columns=[self.S, self.R])
        self.dates = [
            date_obj.strftime(self.DATE_FORMAT) for date_obj in sr_df.index]
        # Minimum size of records
        self.min_size = self.ensure_natural_int(min_size, "min_size")
        # Check length of records
        if self.min_size < 3:
            raise ValueError(
                f"@min_size must be over 2, but {min_size} was applied.")
        if len(self.dates) < self.min_size * 2:
            raise ValueError(
                f"More than {min_size * 2} records must be included.")
        # Minimum value of RMSLE score
        self.max_rmsle = self.ensure_float(max_rmsle)
        # Setting for optimization
        self._change_dates = []

    def run(self):
        """
        Run optimization and find change points.

        Returns:
            covsirphy.ChangeFinder: self
        """
        # Convert the dataset, index: Recovered, column: log10(Susceptible)
        sr_df = self.sr_df.copy()
        sr_df[self.S] = np.log10(sr_df[self.S].astype(np.float64))
        df = sr_df.pivot_table(index=self.R, values=self.S, aggfunc="last")
        # Convert index to serial numbers
        serial_df = pd.DataFrame(np.arange(1, df.index.max() + 1, 1))
        serial_df.index += 1
        df = df.join(serial_df, how="outer")
        series = df.reset_index(drop=True).iloc[:, 0]
        series = series.dropna()
        # Detection with Ruptures
        warnings.simplefilter("ignore", category=RuntimeWarning)
        algorithm = rpt.Pelt(model="rbf", jump=1, min_size=self.min_size)
        results = algorithm.fit_predict(series.values, pen=0.5)
        # Convert index values to Susceptible values
        reset_series = series.reset_index(drop=True)
        reset_series.index += 1
        susceptible_df = reset_series[results].reset_index()
        # Convert Susceptible values to dates
        df = pd.merge_asof(
            susceptible_df.sort_values(self.S),
            sr_df.reset_index().sort_values(self.S),
            on=self.S, direction="nearest"
        )
        found_list = df[self.DATE].sort_values()[:-1]
        # Only use dates when the previous phase has more than {min_size + 1} days
        delta_days = timedelta(days=self.min_size)
        first_obj = self.date_obj(self.dates[0])
        last_obj = self.date_obj(self.dates[-1])
        effective_list = [first_obj]
        for found in found_list:
            if effective_list[-1] + delta_days < found:
                effective_list.append(found)
        # The last change date must be under the last date of records {- min_size} days
        if effective_list[-1] >= last_obj - delta_days:
            effective_list = effective_list[:-1]
        # Set change points
        self._change_dates = [
            date.strftime(self.DATE_FORMAT) for date in effective_list[1:]]
        return self

    def _curve_fitting(self, phase, start_date, end_date):
        """
        Perform curve fitting for the phase.

        Args:
            phase (str): phase name
            start_date (str): start date of the phase
            end_date (str): end date of the phase

        Returns:
            tuple
                (pandas.DataFrame): Result of curve fitting
                    Index: reset index
                    Columns:
                        - (phase name)_predicted: predicted value of Susceptible
                        - (phase_name)_actual: actual value of Susceptible
                        - (phase_name)_Recovered: Recovered
                (int): minimum value of R, which is the change point of the curve
        """
        sr_df = self.sr_df.copy()
        sta = self.date_obj(start_date)
        end = self.date_obj(end_date)
        sr_df = sr_df.loc[(sr_df.index >= sta) & (sr_df.index <= end), :]
        # Calculate linear and negative exponential curve fitting
        # and select the method with the smaller RMSLE score
        trend = Trend(sr_df=sr_df)
        trend.run()
        df = trend.result_df.copy()
        rmsle = trend.rmsle()
        if rmsle > self.max_rmsle:
            df[f"{self.S}{self.P}"] = None
        # Get min value for vline
        r_value = int(df[self.R].min())
        # Rename the columns
        phase = self.INITIAL if phase == "0th" else phase
        df = df.rename({f"{self.S}{self.P}": f"{phase}{self.P}"}, axis=1)
        df = df.rename({f"{self.S}{self.A}": f"{phase}{self.A}"}, axis=1)
        df = df.rename({f"{self.R}": f"{phase}_{self.R}"}, axis=1)
        return (df, r_value)

    def date_range(self, change_dates=None):
        """
        Calculate the list of start/end dates with the list of change dates.

        Args:
            change_dates (list[str] or None): list of change points

        Returns:
            tuple(list[str], list[str]): list of start/end dates

        Notes:
            @change_dates must be specified if ChangeFinder.run() was not done.
        """
        change_dates = change_dates or self._change_dates[:]
        self._change_dates = change_dates[:]
        if not change_dates:
            return([self.dates[0]], [self.dates[-1]])
        # Start dates
        start_dates = [self.dates[0], *change_dates]
        # End dates
        end_dates = [self.yesterday(date) for date in change_dates]
        end_dates.append(self.dates[-1])
        return (start_dates, end_dates)

    def show(self, area, change_dates=None, filename=None):
        """
        show the S-R trend in a figure.

        Args:
            area (str): area name
            change_dates (list[str] or None): list of change points
            filename (str): filename of the figure, or None (display)

        Notes:
            @change_dates must be specified if ChangeFinder.run() was not done.
        """
        # Curve fitting
        start_dates, end_dates = self.date_range(change_dates)
        nested = [
            self._curve_fitting(self.num2str(num), start_date, end_date)
            for (num, (start_date, end_date))
            in enumerate(zip(start_dates, end_dates))
        ]
        df_list, vlines = zip(*nested)
        comp_df = pd.concat([self.sr_df, *df_list], axis=1)
        comp_df = comp_df.rename({self.S: f"{self.S}{self.A}"}, axis=1)
        comp_df = comp_df.apply(
            lambda x: pd.to_numeric(x, errors="coerce", downcast="integer"),
            axis=0
        )
        # Show figure
        pred_cols = [
            col for col in comp_df.columns if col.endswith(self.P)
        ]
        if len(pred_cols) == 1:
            title = f"{area}: S-R trend without change points"
        else:
            _list = self._change_dates[:]
            strings = [
                ", ".join(_list[i: i + 6]) for i in range(0, len(_list), 6)
            ]
            change_str = ",\n".join(strings)
            title = f"{area}: S-R trend changed on\n{change_str}"
        Trend.show_with_many(
            result_df=comp_df,
            predicted_cols=pred_cols,
            title=title,
            vlines=vlines[1:],
            filename=filename
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from covsirphy.cleaning.term import Term


class ModelBase(Term):
    """
    Base class of ODE models.
    """
    # Quartile range of the parametes when setting initial values
    QUANTILE_RANGE = [0.3, 0.7]
    # Model name
    NAME = "ModelBase"
    # names of parameters
    PARAMETERS = list()
    DAY_PARAMETERS = list()
    # Variable names in (non-dim, dimensional) ODEs
    VAR_DICT = dict()
    VARIABLES = list(VAR_DICT.values())
    # Weights of variables in parameter estimation error function
    WEIGHTS = np.array(list())
    # Variables that increases monotonically
    VARS_INCLEASE = list()
    # Example set of parameters and initial values
    EXAMPLE = {
        Term.STEP_N: 180,
        Term.N.lower(): 1_000_000,
        Term.PARAM_DICT: dict(),
        Term.Y0_DICT: dict(),
    }

    def __init__(self, population):
        """
        This method should be overwritten in subclass.

        Args:
            population (int): total population
        """
        # Total population
        self.population = self.ensure_natural_int(
            population, name="population"
        )
        # Dictionary of non-dim parameters: {name: value}
        self.non_param_dict = {}

    def __str__(self):
        param_str = ", ".join(
            [f"{p}={v}" for (p, v) in self.non_param_dict.items()]
        )
        return f"{self.NAME} model with {param_str}"

    def __getitem__(self, key):
        """
        Args:
            key (str): parameter name
        """
        if key not in self.non_param_dict.keys():
            raise KeyError(f"key must be in {', '.join(self.PARAMETERS)}")
        return self.non_param_dict[key]

    def __call__(self, t, X):
        """
        Return the list of dS/dt (tau-free) etc.
        This method should be overwritten in subclass.

        Returns:
            np.array
        """
        raise NotImplementedError

    @classmethod
    def param_range(cls, taufree_df, population):
        """
        Define the range of parameters (not including tau value).
        This method should be overwritten in subclass.

        Args:
            taufree_df (pandas.DataFrame):
                Index:
                    reset index
                Columns:
                    - t (int): time steps (tau-free)
                    - columns with dimensional variables
            population (int): total population

        Returns:
            (dict)
                - key (str): parameter name
                - value (tuple(float, float)): min value and max value
        """
        raise NotImplementedError

    @classmethod
    def specialize(cls, data_df, population):
        """
        Specialize the dataset for this model.
        This method should be overwritten in subclass.

        Args:
            data_df (pandas.DataFrame):
                Index:
                    reset index
                Columns:
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - any columns
            population (int): total population in the place

        Returns:
            (pandas.DataFrame):
                Index:
                    reset index
                Columns:
                    - any columns @data_df has
                    - columns with dimensional variables
        """
        raise NotImplementedError

    @classmethod
    def restore(cls, specialized_df):
        """
        Restore Confirmed/Infected/Recovered/Fatal using a dataframe with the variables of the model.
        This method should be overwritten in subclass.

        Args:
            specialized_df (pandas.DataFrame): dataframe with the variables

                Index:
                    (object):
                Columns:
                    - variables of the models (int)
                    - any columns

        Returns:
            (pandas.DataFrame):
                Index:
                    (object): as-is
                Columns:
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - the other columns @specialzed_df has
        """
        df = specialized_df.copy()
        other_cols = list(set(df.columns) - set(cls.VALUE_COLUMNS))
        df[cls.C] = None
        df[cls.CI] = None
        df[cls.F] = None
        df[cls.R] = None
        return df.loc[:, [*cls.VALUE_COLUMNS, *other_cols]]

    def calc_r0(self):
        """
        Calculate (basic) reproduction number.
        This method should be overwritten in subclass.

        Returns:
            float
        """
        raise NotImplementedError

    def calc_days_dict(self, tau):
        """
        Calculate 1/beta [day] etc.
        This method should be overwritten in subclass.

        Args:
            param tau (int): tau value [min]

        Returns:
            dict[str, int]
        """
        raise NotImplementedError

    @classmethod
    def tau_free(cls, subset_df, population, tau=None):
        """
        Create a dataframe specialized to the model.
        If tau is not None, Date column will be converted to '(Date - start date) / tau'
        and saved in t column.

        Args:
            subset_df (pandas.DataFrame):
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - any columns
            population (int): population value
            tau (int or None): tau value [min], 0 <= tau <= 1440

        Returns:
            (pandas.DataFrame):
                Index:
                    reset index
                Columns:
                    - t (int): if tau is not None
                    - columns with dimensional variables
        """
        df = subset_df.copy()
        if tau is None:
            df = df.drop(cls.DATE, axis=1)
            return cls.specialize(df, population=population).reset_index(drop=True)
        tau = cls.ensure_tau(tau)
        df = cls.ensure_dataframe(
            df, name="data_df", columns=cls.NLOC_COLUMNS)
        # Calculate elapsed time from the first date [min]
        df[cls.T] = (df[cls.DATE] - df[cls.DATE].min()).dt.total_seconds()
        df[cls.T] = df[cls.T] // 60
        # Convert to tau-free
        df[cls.TS] = (df[cls.T] / tau).astype(np.int64)
        df = df.drop([cls.T, cls.DATE], axis=1)
        return cls.specialize(df, population).reset_index(drop=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from covsirphy import DataLoader


@pytest.fixture(scope="session")
def data_loader():
    return DataLoader()


@pytest.fixture(scope="session")
def jhu_data(data_loader):
    return data_loader.jhu()


@pytest.fixture(scope="session")
def population_data(data_loader):
    return data_loader.population()


@pytest.fixture(scope="session")
def oxcgrt_data(data_loader):
    return data_loader.oxcgrt()


@pytest.fixture(scope="session")
def japan_data(data_loader):
    return data_loader.japan()


@pytest.fixture(scope="session")
def linelist_data(data_loader):
    return data_loader.linelist()


@pytest.fixture(scope="session")
def pcr_data(data_loader):
    return data_loader.pcr()


@pytest.fixture(scope="session")
def vaccine_data(data_loader):
    return data_loader.vaccine()


@pytest.fixture(scope="session")
def pyramid_data(data_loader):
    return data_loader.pyramid()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
from datetime import timedelta
from math import log10, floor
import warnings
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn import linear_model
from covsirphy.util.error import deprecate, ScenarioNotFoundError, UnExecutedError
from covsirphy.util.plotting import line_plot, box_plot
from covsirphy.cleaning.oxcgrt import OxCGRTData
from covsirphy.analysis.param_tracker import ParamTracker
from covsirphy.analysis.data_handler import DataHandler


class Scenario(DataHandler):
    """
    Scenario analysis.

    Args:
        jhu_data (covsirphy.JHUData): object of records
        population_data (covsirphy.PopulationData): PopulationData object
        country (str): country name
        province (str or None): province name
        tau (int or None): tau value
        auto_complement (bool): if True and necessary, the number of cases will be complemented
    """

    def __init__(self, jhu_data, population_data, country, province=None, tau=None, auto_complement=True):
        super().__init__(
            jhu_data, population_data, country, province=province, auto_complement=auto_complement)
        self.tau = self.ensure_tau(tau)
        self._update_range(first_date=None, last_date=None)
        # Linear regression modes for prediction of ODE parameters
        self._oxcgrt_data = None
        self._lm_dict = {}

    def __getitem__(self, key):
        """
        Return the phase series object for the scenario.

        Args:
            key (str): scenario name

        Raises:
            ScenarioNotFoundError: the scenario is not registered

        Returns:
            covsirphy.PhaseSeries
        """
        if key in self._tracker_dict:
            return self._tracker_dict[key].series
        raise ScenarioNotFoundError(key)

    def __setitem__(self, key, value):
        """
        Register a phase series.

        Args:
            key (str): scenario name
            value (covsirphy.PhaseSeries): phase series object
        """
        self._tracker_dict[key] = ParamTracker(
            self.record_df, value, area=self.area, tau=self.tau)

    @property
    def first_date(self):
        """
        str: the first date of the records
        """
        return self._first_date

    @first_date.setter
    def first_date(self, date):
        self._update_range(first_date=date)

    @property
    def last_date(self):
        """
        str: the last date of the records
        """
        return self._last_date

    @last_date.setter
    def last_date(self, date):
        self._update_range(last_date=date)

    def _update_range(self, first_date=None, last_date=None):
        """
        Set first/last dates of the records.
        Records to analyse will be updated and all scenarios will be cleared.

        Args:
            first_date (str or None): the first date of the records
            last_date (str or None): the last date of the records
        """
        if first_date is not None:
            self._first_date = self._ensure_date_in_range(first_date)
        if last_date is not None:
            self._last_date = self._ensure_date_in_range(last_date)
        super().init_records()
        self._init_trackers()

    def _ensure_date_in_range(self, date):
        """
        Ensure that the date is in the range of (the first date, the last date).

        Args:
            date (str): date, like 01Jan2020
        """
        self.ensure_date_order(self._first_date, date, name="date")
        self.ensure_date_order(date, self._last_date, name="date")
        return date

    def _init_trackers(self):
        """
        Initialize dictionary of trackers.
        """
        series = ParamTracker.create_series(
            first_date=self._first_date, last_date=self._last_date, population=self.population)
        tracker = ParamTracker(
            record_df=self.record_df, phase_series=series, area=self.area, tau=self.tau)
        self._tracker_dict = {self.MAIN: tracker}

    def _tracker(self, name, template="Main"):
        """
        Ensure that the phases series is registered.
        If not registered, copy the template phase series.

        Args:
            name (str): phase series name
            template (str): name of template phase series

        Returns:
            covsirphy.ParamTracker
        """
        # Registered
        if name in self._tracker_dict:
            return self._tracker_dict[name]
        # Un-registered and create it
        if template not in self._tracker_dict:
            raise ScenarioNotFoundError(template)
        tracker = copy.deepcopy(self._tracker_dict[template])
        self._tracker_dict[name] = tracker
        return tracker

    @deprecate(old="Scenario.add_phase()", new="Scenario.add()")
    def add_phase(self, **kwargs):
        return self.add(**kwargs)

    def add(self, name="Main", end_date=None, days=None,
            population=None, model=None, **kwargs):
        """
        Add a new phase.
        The start date will be the next date of the last registered phase.

        Args:
            name (str): phase series name, 'Main' or user-defined name
            end_date (str): end date of the new phase
            days (int): the number of days to add
            population (int or None): population value of the start date
            model (covsirphy.ModelBase or None): ODE model
            kwargs: keyword arguments of ODE model parameters, not including tau value.

        Returns:
            covsirphy.Scenario: self

        Notes:
            - If the phases series has not been registered, new phase series will be created.
            - Either @end_date or @days must be specified.
            - If @end_date and @days are None, the end date will be the last date of the records.
            - If both of @end_date and @days were specified, @end_date will be used.
            - If @popultion is None, initial value will be used.
            - If @model is None, the model of the last phase will be used.
            - Tau will be fixed as the last phase's value.
            - kwargs: Default values are the parameter values of the last phase.
        """
        if end_date is not None:
            self.ensure_date(end_date, name="end_date")
        tracker = self._tracker(name)
        try:
            tracker.add(
                end_date=end_date, days=days, population=population,
                model=model, **kwargs)
        except ValueError:
            last_date = tracker.series.unit("last").end_date
            raise ValueError(
                f'@end_date must be over {last_date}. However, {end_date} was applied.') from None
        self._tracker_dict[name] = tracker
        return self

    def clear(self, name="Main", include_past=False, template="Main"):
        """
        Clear phase information.

        Args:
            name (str): scenario name
            include_past (bool): if True, past phases will be removed as well as future phases
            template (str): name of template scenario

        Returns:
            covsirphy.Scenario: self

        Notes:
            If un-registered scenario name was specified, new scenario will be created.
            Future phases will be always deleted.
        """
        tracker = self._tracker(name, template=template)
        if include_past:
            self[name] = tracker.delete_all()
        else:
            self[name] = tracker.delete(phases=tracker.future_phases()[0])
        return self

    def _delete_series(self, name):
        """
        Delete a scenario or initialise main scenario.

        Args:
            name (str): name of phase series

        Returns:
            covsirphy.Scenario: self
        """
        if name == self.MAIN:
            self[self.MAIN] = self._tracker(self.MAIN).delete_all()
        else:
            self._tracker_dict.pop(name)
        return self

    def delete(self, phases=None, name="Main"):
        """
        Delete phases.

        Args:
            phase (list[str] or None): phase names, or ['last']
            name (str): name of phase series

        Returns:
            covsirphy.Scenario: self

        Notes:
            If @phases is None, the phase series will be deleted.
            When @phase is '0th', disable 0th phase. 0th phase will not be deleted.
            If the last phase is included in @phases, the dates will be released from phases.
            If the last phase is not included, the dates will be assigned to the previous phase.
        """
        # Clear main series or delete sub phase series
        if phases is None:
            return self._delete_series(name)
        # Delete phases
        tracker = self._tracker(name)
        self[name] = tracker.delete(phases=phases)
        return self

    def disable(self, phases, name="Main"):
        """
        The phases will be disabled and removed from summary.

        Args:
            phase (list[str] or None): phase names
            name (str): name of phase series

        Returns:
            covsirphy.Scenario: self
        """
        self[name] = self._tracker(name).disable(phases)
        return self

    def enable(self, phases, name="Main"):
        """
        The phases will be enabled and appear in summary.

        Args:
            phase (list[str] or None): phase names
            name (str): name of phase series

        Returns:
            covsirphy.Scenario: self
        """
        self[name] = self._tracker(name).enable(phases)
        return self

    def combine(self, phases, name="Main", population=None, **kwargs):
        """
        Combine the sequential phases as one phase.
        New phase name will be automatically determined.

        Args:
            phases (list[str]): list of phases
            name (str, optional): name of phase series
            population (int): population value of the start date
            kwargs: keyword arguments to save as phase information

        Raises:
            TypeError: @phases is not a list

        Returns:
            covsirphy.Scenario: self
        """
        self[name] = self._tracker(name).combine(
            phases=phases, population=population, **kwargs)
        return self

    def separate(self, date, name="Main", population=None, **kwargs):
        """
        Create a new phase with the change point.
        New phase name will be automatically determined.

        Args:
            date (str): change point, i.e. start date of the new phase
            name (str): scenario name
            population (int): population value of the change point
            kwargs: keyword arguments of PhaseUnit.set_ode() if update is necessary

        Returns:
            covsirphy.Scenario: self
        """
        self[name] = self._tracker(name).separate(
            date=date, population=population, **kwargs)
        return self

    def _summary(self, name=None):
        """
        Summarize the series of phases and return a dataframe.

        Args:
            name (str): phase series name
                - name of alternative phase series registered by Scenario.add()
                - if None, all phase series will be shown

        Returns:
            pandas.DataFrame:
            - if @name not None, as the same as PhaseSeries().summary()
            - if @name is None, index will be phase series name and phase name

        Notes:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        if name is None:
            if len(self._tracker_dict.keys()) > 1:
                dataframes = []
                for (_name, tracker) in self._tracker_dict.items():
                    summary_df = tracker.series.summary()
                    summary_df = summary_df.rename_axis(self.PHASE)
                    summary_df[self.SERIES] = _name
                    dataframes.append(summary_df.reset_index())
                df = pd.concat(dataframes, ignore_index=True, sort=False)
                return df.set_index([self.SERIES, self.PHASE])
            name = self.MAIN
        return self._tracker(name).series.summary()

    def summary(self, columns=None, name=None):
        """
        Summarize the series of phases and return a dataframe.

        Args:
            name (str): phase series name
                - name of alternative phase series registered by Scenario.add()
                - if None, all phase series will be shown
            columns (list[str] or None): columns to show

        Returns:
            pandas.DataFrame:
            - if @name not None, as the same as PhaseSeries().summary()
            - if @name is None, index will be phase series name and phase name

        Notes:
            If 'Main' was used as @name, main PhaseSeries will be used.
            If @columns is None, all columns will be shown.
        """
        df = self._summary(name=name)
        all_cols = df.columns.tolist()
        if set(self.EST_COLS).issubset(all_cols):
            all_cols = [col for col in all_cols if col not in self.EST_COLS]
            all_cols += self.EST_COLS
        columns = columns or all_cols
        self.ensure_list(columns, candidates=all_cols, name="columns")
        df = df.loc[:, columns]
        return df.dropna(how="all", axis=1).fillna(self.UNKNOWN)

    def trend(self, force=True, name="Main", show_figure=True, filename=None, **kwargs):
        """
        Perform S-R trend analysis and set phases.

        Args:
            force (bool): if True, change points will be over-written
            name (str): phase series name
            show_figure (bool): if True, show the result as a figure
            filename (str): filename of the figure, or None (display)
            kwargs: keyword arguments of ChangeFinder()

        Returns:
            covsirphy.Scenario: self
        """
        # Arguments
        if "n_points" in kwargs.keys():
            raise ValueError(
                "@n_points argument is un-necessary"
                " because the number of change points will be automatically determined."
            )
        try:
            include_init_phase = kwargs.pop("include_init_phase")
            warnings.warn(
                "@include_init_phase was deprecated. Please use Scenario.disable('0th').",
                DeprecationWarning, stacklevel=2)
        except KeyError:
            include_init_phase = True
        try:
            force = kwargs.pop("set_phases")
        except KeyError:
            pass
        # S-R trend analysis
        tracker = self._tracker(name)
        self[name] = tracker.trend(
            force=force, show_figure=show_figure, filename=filename, **kwargs)
        # Disable 0th phase, if necessary
        if not include_init_phase:
            self[name] = tracker.disable(phases=["0th"])
        return self

    def estimate(self, model, phases=None, name="Main", n_jobs=-1, **kwargs):
        """
        Perform parameter estimation for each phases.

        Args:
            model (covsirphy.ModelBase): ODE model
            phases (list[str]): list of phase names, like 1st, 2nd...
            name (str): phase series name
            n_jobs (int): the number of parallel jobs or -1 (CPU count)
            kwargs: keyword arguments of model parameters and covsirphy.Estimator.run()

        Notes:
            - If 'Main' was used as @name, main PhaseSeries will be used.
            - If @name phase was not registered, new PhaseSeries will be created.
            - If @phases is None, all past phase will be used.
            - Phases with estimated parameter values will be ignored.
            - In kwargs, tau value cannot be included.
        """
        if self.TAU in kwargs:
            raise ValueError(
                "@tau must be specified when scenario = Scenario(), and cannot be specified here.")
        self.tau, self[name] = self._tracker(name).estimate(
            model=model, phases=phases, n_jobs=n_jobs, **kwargs)

    def phase_estimator(self, phase, name="Main"):
        """
        Return the estimator of the phase.

        Args:
            phase (str): phase name, like 1st, 2nd...
            name (str): phase series name

        Return:
            covsirphy.Estimator: estimator of the phase
        """
        estimator = self._tracker_dict[name].series.unit(phase).estimator
        if estimator is None:
            raise UnExecutedError(
                f'Scenario.estimate(model, phases=["{phase}"], name={name})')
        return estimator

    def estimate_history(self, phase, name="Main", **kwargs):
        """
        Show the history of optimization.

        Args:
            phase (str): phase name, like 1st, 2nd...
            name (str): phase series name
            kwargs: keyword arguments of covsirphy.Estimator.history()

        Notes:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        estimator = self.phase_estimator(phase=phase, name=name)
        estimator.history(**kwargs)

    def estimate_accuracy(self, phase, name="Main", **kwargs):
        """
        Show the accuracy as a figure.

        Args:
            phase (str): phase name, like 1st, 2nd...
            name (str): phase series name
            kwargs: keyword arguments of covsirphy.Estimator.accuracy()

        Notes:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        estimator = self.phase_estimator(phase=phase, name=name)
        estimator.accuracy(**kwargs)

    def simulate(self, name="Main", y0_dict=None, show_figure=True, filename=None):
        """
        Simulate ODE models with set/estimated parameter values and show it as a figure.

        Args:
            name (str): phase series name. If 'Main', main PhaseSeries will be used
            y0_dict(dict[str, float] or None): dictionary of initial values of variables
            show_figure (bool): if True, show the result as a figure
            filename (str): filename of the figure, or None (show figure)

        Returns:
            pandas.DataFrame
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Country (str): country/region name
                    - Province (str): province/prefecture/state name
                    - Variables of the model and dataset (int): Confirmed etc.
        """
        tracker = self._tracker(name)
        try:
            sim_df = tracker.simulate(y0_dict=y0_dict)
        except UnExecutedError:
            raise UnExecutedError(
                "Scenario.trend() or Scenario.add(), and Scenario.estimate(model)") from None
        if not show_figure:
            return sim_df
        # Show figure
        df = sim_df.set_index(self.DATE)
        fig_cols_set = set(df.columns) & set(self.FIG_COLUMNS)
        fig_cols = [col for col in self.FIG_COLUMNS if col in fig_cols_set]
        line_plot(
            df[fig_cols],
            title=f"{self.area}: Simulated number of cases ({name} scenario)",
            filename=filename,
            y_integer=True,
            v=tracker.change_dates()
        )
        return sim_df

    def get(self, param, phase="last", name="Main"):
        """
        Get the parameter value of the phase.

        Args:
            param (str): parameter name (columns in self.summary())
            phase (str): phase name or 'last'
                - if 'last', the value of the last phase will be returned
            name (str): phase series name

        Returns:
            str or int or float

        Notes:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        df = self.summary(name=name)
        if param not in df.columns:
            raise KeyError(f"@param must be in {', '.join(df.columns)}.")
        if phase == "last":
            phase = df.index[-1]
        return df.loc[phase, param]

    def _param_history(self, targets, name):
        """
        Return the subset of summary dataframe to select the target of parameter history.

        Args:
            targets (list[str] or str): parameters to show (Rt etc.)
            name (str): phase series name

        Returns:
            pandas.DataFrame: selected summary dataframe

        Raises:
            KeyError: targets are not in the columns of summary dataframe
        """
        series = self._tracker_dict[name].series
        model_set = {unit.model for unit in series}
        model_set = model_set - set([None])
        parameters = self.flatten([m.PARAMETERS for m in model_set])
        day_params = self.flatten([m.DAY_PARAMETERS for m in model_set])
        selectable_cols = [self.N, *parameters, self.RT, *day_params]
        selectable_set = set(selectable_cols)
        df = series.summary().replace(self.UNKNOWN, None)
        if not selectable_set.issubset(df.columns):
            raise UnExecutedError(
                f'Scenario.estimate(model, phases=None, name="{name}")')
        targets = [targets] if isinstance(targets, str) else targets
        targets = targets or selectable_cols
        if not set(targets).issubset(selectable_set):
            raise KeyError(
                f"@targets must be selected from {', '.join(selectable_cols)}."
            )
        df = df.loc[:, targets].dropna(how="any", axis=0)
        return df.astype(np.float64)

    @deprecate(
        old="Scenario.param_history(targets: list)",
        new="Scenario.history(target: str)",
        version="2.7.3-alpha")
    def param_history(self, targets=None, name="Main", divide_by_first=True,
                      show_figure=True, filename=None, show_box_plot=True, **kwargs):
        """
        Return subset of summary and show a figure to show the history.

        Args:
            targets (list[str] or str): parameters to show (Rt etc.)
            name (str): phase series name
            divide_by_first (bool): if True, divide the values by 1st phase's values
            show_box_plot (bool): if True, box plot. if False, line plot
            show_figure (bool): If True, show the result as a figure
            filename (str): filename of the figure, or None (show figure)
            kwargs: keyword arguments of pd.DataFrame.plot or line_plot()

        Returns:
            pandas.DataFrame

        Notes:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        self._tracker(name)
        # Select target to show
        df = self._param_history(targets, name)
        # Divide by the first phase parameters
        if divide_by_first:
            df = df / df.iloc[0, :]
            title = f"{self.area}: Ratio to 1st phase parameters ({name} scenario)"
        else:
            title = f"{self.area}: History of parameter values ({name} scenario)"
        if not show_figure:
            return df
        if show_box_plot:
            h_values = [1.0] if divide_by_first or self.RT in targets else None
            box_plot(df, title, h=h_values, filename=filename)
            return df
        _df = df.reset_index(drop=True)
        _df.index = _df.index + 1
        h = 1.0 if divide_by_first else None
        line_plot(
            _df, title=title,
            xlabel="Phase", ylabel=str(), math_scale=False, h=h,
            filename=filename
        )
        return df

    def _describe(self, y0_dict=None):
        """
        Describe representative values.

        Args:
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value

        Returns:
            pandas.DataFrame
                Index:
                    (int): scenario name
                Columns:
                    - max(Infected): max value of Infected
                    - argmax(Infected): the date when Infected shows max value
                    - Confirmed({date}): Confirmed on the next date of the last phase
                    - Infected({date}): Infected on the next date of the last phase
                    - Fatal({date}): Fatal on the next date of the last phase
        """
        _dict = {}
        for (name, _) in self._tracker_dict.items():
            # Predict the number of cases
            df = self.simulate(name=name, y0_dict=y0_dict, show_figure=False)
            df = df.set_index(self.DATE)
            cols = df.columns[:]
            last_date = df.index[-1]
            # Max value of Infected
            max_ci = df[self.CI].max()
            argmax_ci = df[self.CI].idxmax().strftime(self.DATE_FORMAT)
            # Confirmed on the next date of the last phase
            last_c = df.loc[last_date, self.C]
            # Infected on the next date of the last phase
            last_ci = df.loc[last_date, self.CI]
            # Fatal on the next date of the last phase
            last_f = df.loc[last_date, self.F] if self.F in cols else None
            # Save representative values
            last_date_str = last_date.strftime(self.DATE_FORMAT)
            _dict[name] = {
                f"max({self.CI})": max_ci,
                f"argmax({self.CI})": argmax_ci,
                f"{self.C} on {last_date_str}": last_c,
                f"{self.CI} on {last_date_str}": last_ci,
                f"{self.F} on {last_date_str}": last_f,
            }
        return pd.DataFrame.from_dict(_dict, orient="index")

    def describe(self, y0_dict=None, with_rt=True):
        """
        Describe representative values.

        Args:
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value
            with_rt (bool): whether show the history of Rt values

        Returns:
            pandas.DataFrame:
                Index:
                    str: scenario name
                Columns:
                    - max(Infected): max value of Infected
                    - argmax(Infected): the date when Infected shows max value
                    - Confirmed({date}): Confirmed on the next date of the last phase
                    - Infected({date}): Infected on the next date of the last phase
                    - Fatal({date}): Fatal on the next date of the last phase
                    - nth_Rt etc.: Rt value if the values are not the same values
        """
        df = self._describe(y0_dict=y0_dict)
        if not with_rt or len(self._tracker_dict) == 1:
            return df
        # History of reproduction number
        rt_df = self.summary().reset_index()
        rt_df = rt_df.pivot_table(
            index=self.SERIES, columns=self.PHASE, values=self.RT)
        rt_df = rt_df.fillna(self.UNKNOWN)
        rt_df = rt_df.loc[:, rt_df.nunique() > 1]
        cols = sorted(rt_df, key=self.str2num)
        return df.join(rt_df[cols].add_suffix(f"_{self.RT}"), how="left")

    def _track_param(self, name):
        """
        Get the history of parameters for the scenario.

        Args:
            name (str): phase series name

        Returns:
            pandas.DataFrame:
                Index: Date (pandas.TimeStamp)
                Columns:
                    - Population (int)
                    - Rt (float)
                    - parameter values (float)
                    - day parameter values (float)
        """
        df = self.summary(name=name).replace(self.UNKNOWN, None)
        # Date range to dates
        df[self.START] = pd.to_datetime(df[self.START])
        df[self.END] = pd.to_datetime(df[self.END])
        df[self.DATE] = df[[self.START, self.END]].apply(
            lambda x: pd.date_range(x[0], x[1]).tolist(), axis=1)
        df = df.reset_index(drop=True).explode(self.DATE)
        # Columns
        df = df.drop(
            [self.TENSE, self.START, self.END, self.ODE, self.TAU, *self.EST_COLS],
            axis=1, errors="ignore")
        df = df.set_index(self.DATE)
        for col in df.columns:
            df[col] = pd.to_numeric(df[col], errors="coerce")
        df[self.N] = df[self.N].astype(np.int64)
        return df

    def _track(self, name, y0_dict=None):
        """
        Show values of parameters and variables in one dataframe for the scenario.

        Args:
            name (str): phase series name
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value

        Returns:
            pandas.DataFrame:
                Index: reset index
                Columns:
                    - Date (pandas.TimeStamp)
                    - variables (int)
                    - Population (int)
                    - Rt (float)
                    - parameter values (float)
                    - day parameter values (float)
        """
        sim_df = self.simulate(name=name, y0_dict=y0_dict, show_figure=False)
        param_df = self._track_param(name=name)
        return pd.merge(
            sim_df, param_df, how="inner",
            left_on=self.DATE, right_index=True, sort=True
        )

    def track(self, with_actual=True, y0_dict=None):
        """
        Show values of parameters and variables in one dataframe.

        Args:
            with_actual (bool): if True, show actual number of cases will included as "Actual" scenario
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value

        Returns:
            pandas.DataFrame:
                Index: reset index
                Columns:
                    - Scenario (str)
                    - Date (pandas.TimeStamp)
                    - variables (int)
                    - Population (int)
                    - Rt (float)
                    - parameter values (float)
                    - day parameter values (float)
        """
        dataframes = []
        append = dataframes.append
        for name in self._tracker_dict.keys():
            df = self._track(name=name, y0_dict=y0_dict)
            df.insert(0, self.SERIES, name)
            append(df)
        if with_actual:
            df = self.records(show_figure=False)
            df.insert(0, self.SERIES, self.ACTUAL)
            append(df)
        return pd.concat(dataframes, axis=0, sort=False)

    def _history(self, target, with_actual=True, y0_dict=None):
        """
        Show the history of variables and parameter values to compare scenarios.

        Args:
            target (str): parameter or variable name to show (Rt, Infected etc.)
            with_actual (bool): if True and @target is a variable name, show actual number of cases
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value

        Returns:
            pandas.DataFrame
        """
        # Include actual data or not
        with_actual = with_actual and target in self.VALUE_COLUMNS
        # Get tracking data
        df = self.track(with_actual=with_actual, y0_dict=y0_dict)
        if target not in df.columns:
            col_str = ", ".join(list(df.columns))
            raise KeyError(
                f"@target must be selected from {col_str}, but {target} was applied.")
        # Select the records of target variable
        return df.pivot_table(
            values=target, index=self.DATE, columns=self.SERIES, aggfunc="last")

    def history(self, target, with_actual=True, y0_dict=None, show_figure=True, filename=None):
        """
        Show the history of variables and parameter values to compare scenarios.

        Args:
            target (str): parameter or variable name to show (Rt, Infected etc.)
            with_actual (bool): if True and @target is a variable name, show actual number of cases
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value
            show_figure (bool): If True, show the result as a figure
            filename (str): filename of the figure, or None (show figure)

        Returns:
            pandas.DataFrame
        """
        df = self._history(
            target=target, with_actual=with_actual, y0_dict=y0_dict)
        if not show_figure:
            return df
        if target == self.RT:
            ylabel = self.RT_FULL
        elif target in self.VALUE_COLUMNS:
            ylabel = f"The number of {target.lower()} cases"
        else:
            ylabel = target
        title = f"{self.area}: {ylabel} over time"
        tracker = self._tracker(self.MAIN)
        line_plot(
            df, title, ylabel=ylabel,
            h=1.0 if target == self.RT else None,
            v=tracker.change_dates(),
            math_scale=False,
            filename=filename
        )
        return df

    def history_rate(self, params=None, name="Main", show_figure=True, filename=None):
        """
        Show change rates of parameter values in one figure.
        We can find the parameters which increased/decreased significantly.

        Args:
            params (list[str] or None): parameters to show
            name (str): phase series name
            show_figure (bool): If True, show the result as a figure
            filename (str): filename of the figure, or None (show figure)

        Returns:
            pandas.DataFrame
        """
        df = self._track_param(name=name)
        model = self._tracker(name).last_model
        cols = list(set(df.columns) & set(model.PARAMETERS))
        if params is not None:
            if not isinstance(params, (list, set)):
                raise TypeError(
                    f"@params must be a list of parameters, but {params} were applied.")
            cols = list(set(cols) & set(params)) or cols
        df = df.loc[:, cols] / df.loc[df.index[0], cols]
        if show_figure:
            f_date = df.index[0].strftime(self.DATE_FORMAT)
            title = f"{self.area}: {model.NAME} parameter change rates over time (1.0 on {f_date})"
            ylabel = f"Value per that on {f_date}"
            title = f"{self.area}: {ylabel} over time"
            tracker = self._tracker(self.MAIN)
            line_plot(
                df, title, ylabel=ylabel, v=tracker.change_dates(),
                math_scale=False, filename=filename)
        return df

    def retrospective(self, beginning_date, model, control="Main", target="Target", **kwargs):
        """
        Perform retrospective analysis.
        Compare the actual series of phases (control) and
        series of phases with specified parameters (target).

        Args:
            beginning_date (str): when the parameter values start to be changed from actual values
            model (covsirphy.ModelBase): ODE model
            control (str): scenario name of control
            target (str): scenario name of target
            kwargs: keyword argument of parameter values and Estimator.run()

        Notes:
            When parameter values are not specified,
            actual values of the last date before the beginning date will be used.
        """
        param_dict = {
            k: v for (k, v) in kwargs.items() if k in model.PARAMETERS}
        est_kwargs = dict(kwargs.items() - param_dict.items())
        # Control
        self.clear(name=control, include_past=True)
        self.trend(name=control, show_figure=False)
        try:
            self.separate(date=beginning_date, name=control)
        except ValueError:
            pass
        self.estimate(model, name=control, **est_kwargs)
        # Target
        self.clear(name=target, include_past=False, template=control)
        phases_changed = [
            self.num2str(i) for (i, ph) in enumerate(self._tracker(target).series)
            if ph >= beginning_date
        ]
        self.delete(phases=phases_changed, name=target)
        self.add(name=target, **param_dict)
        self.estimate(model, name=target, **est_kwargs)

    def score(self, metrics="RMSLE", variables=None, phases=None, past_days=None, name="Main", y0_dict=None):
        """
        Evaluate accuracy of phase setting and parameter estimation of all enabled phases all some past days.

        Args:
            metrics (str): "MAE", "MSE", "MSLE", "RMSE" or "RMSLE"
            variables (list[str] or None): variables to use in calculation
            phases (list[str] or None): phases to use in calculation
            past_days (int or None): how many past days to use in calculation, natural integer
            name(str): phase series name. If 'Main', main PhaseSeries will be used
            y0_dict(dict[str, float] or None): dictionary of initial values of variables

        Returns:
            float: score with the specified metrics

        Notes:
            If @variables is None, ["Infected", "Fatal", "Recovered"] will be used.
            "Confirmed", "Infected", "Fatal" and "Recovered" can be used in @variables.
            If @phases is None, all phases will be used.
            @phases and @past_days can not be specified at the same time.
        """
        tracker = self._tracker(name)
        if past_days is not None:
            if phases is not None:
                raise ValueError(
                    "@phases and @past_days cannot be specified at the same time.")
            past_days = self.ensure_natural_int(past_days, name="past_days")
            # Separate a phase, if possible
            beginning_date = self.date_change(
                self._last_date, days=0 - past_days)
            try:
                tracker.separate(date=beginning_date)
            except ValueError:
                pass
            # Ge the list of target phases
            phases = [
                self.num2str(num) for (num, unit)
                in enumerate(tracker.series)
                if unit >= beginning_date
            ]
        return tracker.score(
            metrics=metrics, variables=variables, phases=phases, y0_dict=y0_dict)

    def _fit_create_data(self, oxcgrt_data, model, name):
        """
        Create train/test dataset for Elastic Net regression,
        assuming that OxCGRT scores will impact on ODE parameter values with delay (recovery period).

        Args:
            oxcgrt_data (covsirphy.OxCGRTData): OxCGRT dataset
            model (covsirphy.ModelBase): ODE model
            name (str): scenario name

        Returns:
            tuple(pandas.DataFrame):
                - X dataset for linear regression
                - y dataset for linear regression
                - X dataset of the target dates
        """
        # Parameter values
        param_df = self._track_param(name=name)[model.PARAMETERS]
        # OxCGRT scores
        oxcgrt_df = oxcgrt_data.subset(
            country=self.country).set_index(self.DATE)[OxCGRTData.OXCGRT_VARS_INDICATORS]
        # Apply delay on OxCGRT data
        delay = self.jhu_data.recovery_period
        oxcgrt_df.index += timedelta(days=delay)
        # Create training/test dataset
        df = param_df.join(oxcgrt_df, how="inner")
        df = df.rolling(window=delay).mean()
        df = df.dropna().drop_duplicates()
        X = df.drop(model.PARAMETERS, axis=1)
        y = df.loc[:, model.PARAMETERS]
        # X dataset of the target dates
        dates = pd.date_range(
            start=param_df.index.max() + timedelta(days=1),
            end=oxcgrt_df.index.max(),
            freq="D")
        X_target = oxcgrt_df.loc[dates]
        return (X, y, X_target)

    def fit(self, oxcgrt_data, name="Main", test_size=0.2, seed=0):
        """
        Learn the relationship of ODE parameter values and delayed OxCGRT scores using Elastic Net regression,
        assuming that OxCGRT scores will impact on ODE parameter values with delay (=recovery period).

        Args:
            oxcgrt_data (covsirphy.OxCGRTData): OxCGRT dataset
            name (str): scenario name
            test_size (float): proportion of the test dataset of Elastic Net regression
            seed (int): random seed when spliting the dataset to train/test data

        Returns:
            dict(str, object):
                - alpha (float): alpha value used in Elastic Net regression
                - l1_ratio (float): l1_ratio value used in Elastic Net regression
                - score_train (float): determination coefficient of train dataset
                - score_test (float): determination coefficient of test dataset
                - intercept (pandas.DataFrame): intercept values (Index: ODE parameters, Columns: OxCGRT indicators)

        Raises:
            UnExecutedError: Scenario.estimate() or Scenario.add() were not performed
        """
        # Register OxCGRT data
        self._oxcgrt_data = self.ensure_instance(
            oxcgrt_data, OxCGRTData, name="oxcgrt_data")
        # ODE model
        model = self._tracker(name).last_model
        if model is None:
            raise UnExecutedError(
                "Scenario.estimate() or Scenario.add()",
                message=f", specifying @model (covsirphy.SIRF etc.) and @name='{name}'.")
        # Create training/test dataset
        X, y, *_ = self._fit_create_data(
            oxcgrt_data=oxcgrt_data, model=model, name=name)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=test_size, random_state=seed)
        # Optimize parameters of Elastic Net regression
        cv = linear_model.MultiTaskElasticNetCV(
            alphas=[0, 0.001, 0.01, 0.1, 1, 10, 100, 1000],
            l1_ratio=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            cv=5, n_jobs=-1)
        cv.fit(X_train, y_train)
        info_dict = {"alpha": cv.alpha_, "l1_ratio": cv.l1_ratio_}
        # Perform Elastic Net (Ridge & Lasso) regression
        lm = linear_model.ElasticNet(**info_dict)
        lm.fit(X_train, y_train)
        # Register the regression model to self
        self._lm_dict[name] = lm
        # Return information regarding regression model
        info_dict.update(
            {
                "score_train": lm.score(X_train, y_train),
                "score_test": lm.score(X_test, y_test),
                "intercept": pd.DataFrame(
                    lm.coef_, index=y_train.columns, columns=X_train.columns),
            }
        )
        return info_dict

    def predict(self, name="Main"):
        """
        Predict parameter values of the future phases using Elastic Net regression with OxCGRT scores,
        assuming that OxCGRT scores will impact on ODE parameter values with delay (=recovery period).
        New future phases will be added (over-written).

        Args:
            name (str): scenario name

        Returns:
            covsirphy.Scenario: self

        Raises:
            UnExecutedError: Scenario.fit() was not performed
        """
        # Arguments
        if name not in self._lm_dict:
            raise UnExecutedError(f"Scenario.fit(oxcgrt_data, name={name})")
        model = self._tracker(name).last_model
        lm = self._lm_dict[name]
        # Clear future phases
        self.clear(name=name)
        # Create X dataset of the target dates
        *_, X_target = self._fit_create_data(
            oxcgrt_data=self._oxcgrt_data, model=model, name=name)
        # -> end_date/parameter values
        df = pd.DataFrame(
            lm.predict(X_target), index=X_target.index, columns=model.PARAMETERS)
        df = df.applymap(
            lambda x: np.around(x, 4 - int(floor(log10(abs(x)))) - 1))
        df.index = [
            date.strftime(self.DATE_FORMAT) for date in df.index]
        df.index.name = "end_date"
        phase_df = df.drop_duplicates(keep="last").reset_index()
        # Set new future phases
        for phase_dict in phase_df.to_dict(orient="records"):
            self.add(name=name, **phase_dict)
        return self

    def fit_predict(self, oxcgrt_data, name="Main", **kwargs):
        """
        Predict parameter values of the future phases using Elastic Net regression with OxCGRT scores,
        assuming that OxCGRT scores will impact on ODE parameter values with delay (=recovery period).
        New future phases will be added (over-written).

        Args:
            oxcgrt_data (covsirphy.OxCGRTData): OxCGRT dataset
            name (str): scenario name
            kwargs: the other arguments of Scenario.fit()

        Returns:
            covsirphy.Scenario: self

        Raises:
            UnExecutedError: Scenario.estimate() or Scenario.add() were not performed
        """
        self.fit(oxcgrt_data=oxcgrt_data, name=name, **kwargs)
        self.predict(name=name)
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from itertools import chain
import numpy as np
import pandas as pd
import swifter
from covsirphy.cleaning.term import Term
from covsirphy.ode.mbase import ModelBase
from covsirphy.cleaning.population import PopulationData
from covsirphy.phase.phase_estimator import MPEstimator
from covsirphy.analysis.scenario import Scenario
from covsirphy.analysis.example_data import ExampleData


class ModelValidator(Term):
    """
    Validate a ODE model as follows.
    - Create a model parameter set, phase length and initial values randomly.
    - Simulate the number of cases with an user-defined parameter set.
    - Perform parameter estimation with the theoretical data (the dataset created with step 1).
    - Compare the user-defined parameter set and estimated parameter set.
    - Repeat with anather set of parameters.
    - Small difference of the two parameter sets means that the model can be used for parameter estimation.

    Args:
        tau (int): tau value [min]
        n_trials (int): the number of trials
        seed (int): random seed

    Notes:
        Population value and initial values are defined by model.EXAMPLE.
        Estimators know tau values before parameter estimation.
    """

    def __init__(self, tau=1440, n_trials=8, seed=0):
        self.n_trials = self.ensure_natural_int(n_trials, name="n_trials")
        self.seed = self.ensure_natural_int(
            seed, name="seed", include_zero=True)
        self.tau = self.ensure_tau(tau)
        # Validated models
        self.model_names = []
        # Dataframes of results created by ._get_result()
        self._results = []
        # To avoid "imported but unused"
        self.__swifter = swifter

    def run(self, model, timeout=180, allowance=(0.98, 1.02), n_jobs=-1):
        """
        Execute model validation.

        Args:
            model (covsirphy.ModelBase): ODE model
            timeout (int): time-out of run
            allowance (tuple(float, float)): the allowance of the predicted value
            n_jobs (int): the number of parallel jobs or -1 (CPU count)

        Returns:
            covsirphy.ModelValidator: self
        """
        model = self.ensure_subclass(model, ModelBase, name="model")
        if model.NAME in self.model_names:
            raise ValueError(f"{model.NAME} has been validated.")
        self.model_names.append(model.NAME)
        # Setup: create parameter set, phase length and processor
        df = self._setup(model, n_trials=self.n_trials, seed=self.seed)
        processor = self._processor(model, df)
        # Parameter estimation
        units_estimated = processor.run(
            timeout=timeout, allowance=allowance, n_jobs=n_jobs)
        # Get estimated parameters
        self._results.append(self._get_result(model, df, units_estimated))
        return self

    def _setup(self, model, n_trials, seed):
        """
        Create a model parameter set, phase length randomly.

        Args:
            model (covsirphy.ModelBase): ODE model
            n_trials (int): the number of trials
            seed (int): random seed

        Returns:
            pandas.DataFrame:
                Index: reset index
                Columns:
                    - (float): parameter values from 0 to 1.0
                    - Rt (float): reproduction number
                    - step_n (int): step number of simulation
        """
        np.random.seed(seed)
        population = model.EXAMPLE[self.N.lower()]
        # Parameter set
        parameters = model.PARAMETERS[:]
        df = pd.DataFrame(
            np.random.rand(n_trials, len(parameters)), columns=parameters)
        # Reproduction number
        df[self.RT] = df.swifter.progress_bar(False).apply(
            lambda x: model(population, **x.to_dict()).calc_r0(),
            axis=1)
        # Tau value
        df[self.TAU] = self.tau
        # Step number
        df[self.STEP_N] = np.random.randint(5, 60, n_trials)
        # Return the setting
        return df

    def _processor(self, model, setting_df):
        """
        Generate multi-processor for parameter estimation,
        registering theoretical data and phase units.

        Args:
            model (covsirphy.ModelBase): ODE model
            setting_df (pandas.DataFrame):
                Index: reset index
                Columns:
                    - (float): parameter values from 0 to 1.0
                    - Rt (float): reproduction number
                    - step_n (int): step number of simulation

        Returns:
            covsirphy.MPEstimator: multi-processor for parameter estimation
        """
        units = []
        # Instance to save theoretical data
        example_data = ExampleData(tau=self.tau, start_date="01Jan2020")
        # Population values
        population = model.EXAMPLE[self.N.lower()]
        population_data = PopulationData(filename=None)
        # Register data for each setting
        for (i, setting_dict) in enumerate(setting_df.to_dict(orient="records")):
            name = f"{model.NAME}_{i}"
            step_n = setting_dict[self.STEP_N]
            param_dict = {
                k: v for (k, v) in setting_dict.items() if k in model.PARAMETERS}
            # Add theoretical data
            example_data.add(
                model, step_n=step_n, country=name, param_dict=param_dict)
            # Population
            population_data.update(population, country=name)
            # Phase unit
            snl = Scenario(
                example_data, population_data, country=name, auto_complement=False)
            snl.add()
            unit = snl[self.MAIN].unit("last").del_id().set_id(country=name)
            units.append(unit)
        # Multi-processor for parameter estimation
        processor = MPEstimator(
            model, jhu_data=example_data, population_data=population_data, tau=self.tau)
        return processor.add(units)

    def _get_result(self, model, setting_df, units_estimated):
        """
        Show the result as a dataframe.

        Args:
            model (covsirphy.ModelBase): ODE model
            setting_df (pandas.DataFrame):
                Index: reset index
                Columns:
                    - (float): parameter values from 0 to 1.0
                    - Rt (float): reproduction number
                    - step_n (int): step number of simulation
            units_estimated (list[covsirphy.PhaseUnit]): phase units with estimated parameters

        Returns:
            pandas.DataFrame:
                Index: reset index
                Columns:
                    - ID (str): ID, like SIR_0
                    - ODE (str): model name
                    - Rt (float): reproduction number set by ._setup() method
                    - Rt_est (float): estimated reproduction number
                    - rho etc. (float): parameter values set by ._setup() method
                    - rho_est etc. (float): estimated parameter values
                    - step_n (int): step number of simulation
                    - RMSLE (float): RMSLE score of parameter estimation
                    - Trials (int): the number of trials in parameter estimation
                    - Runtime (str): runtime of parameter estimation
        """
        df = setting_df.copy()
        df["results"] = [unit.to_dict() for unit in units_estimated]
        df = df.join(
            df["results"].apply(pd.Series), how="left", rsuffix="_est")
        df = df.drop(["results", *model.DAY_PARAMETERS], axis=1)
        cols_to_alternate = [self.RT, *model.PARAMETERS]
        cols_alternated = chain.from_iterable(
            zip(cols_to_alternate, [f"{col}_est" for col in cols_to_alternate]))
        df["ID"] = df[self.ODE].str.cat(df.index.astype("str"), sep="_")
        columns = [
            "ID", self.ODE, *cols_alternated, self.STEP_N, *self.EST_COLS]
        return df.loc[:, columns]

    def summary(self):
        """
        Show the summary of validation.

        Returns:
            pandas.DataFrame:
                Index: reset index
                Columns:
                    - ID (str): ID, like SIR_0
                    - ODE (str): model name
                    - Rt (float): reproduction number set by ._setup() method
                    - Rt_est (float): estimated reproduction number
                    - rho etc. (float): parameter values set by ._setup() method
                    - rho_est etc. (float): estimated parameter values
                    - step_n (int): step number of simulation
                    - RMSLE (float): RMSLE score of parameter estimation
                    - Trials (int): the number of trials in parameter estimation
                    - Runtime (str): runtime of parameter estimation
        """
        df = pd.concat(self._results, ignore_index=True, sort=True)
        pre_cols = ["ID", self.ODE, self.RT, f"{self.RT}_est"]
        post_cols = [self.STEP_N, *self.EST_COLS]
        centers = list(set(df.columns) - set(pre_cols) - set(post_cols))
        centers_sorted = sorted(
            centers, key=lambda x: df.columns.tolist().index(x))
        return df[[*pre_cols, *centers_sorted, *post_cols]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime, timezone, timedelta
from pathlib import Path
from covsirphy.cleaning.term import Term
from covsirphy.cleaning.jhu_data import JHUData
from covsirphy.cleaning.japan_data import JapanData
from covsirphy.cleaning.oxcgrt import OxCGRTData
from covsirphy.cleaning.population import PopulationData
from covsirphy.cleaning.pyramid import PopulationPyramidData
from covsirphy.cleaning.covid19datahub import COVID19DataHub
from covsirphy.cleaning.linelist import LinelistData
from covsirphy.cleaning.pcr_data import PCRData
from covsirphy.cleaning.vaccine_data import VaccineData


class DataLoader(Term):
    """
    Download the dataset and perform data cleaning.

    Args:
        directory (str or pathlib.Path): directory to save the downloaded datasets
        update_interval (int): update interval of the local datasets

    Notes:
        GitHub datasets will be always updated because headers of GET response
        does not have 'Last-Modified' keys.
        If @update_interval hours have passed since the last update of local datasets,
        updating will be forced when updating is not prevented by the methods.

    Examples:
        >>> # Setup
        >>> import covsirphy as cs
        >>> data_loader = cs.DataLoader("input")
        >>> # JHU data: the number of cases
        >>> jhu_data = data_loader.jhu()
        >>> print(jhu_data.citation)
        ...
        >>> print(type(jhu_data.cleaned()))
        <class 'pandas.core.frame.DataFrame'>
        >>> # The number of cases in Japan
        >>> jpn_data = data_loader.japan()
        >>> print(jpn_data.citation)
        ...
        >>> print(type(jpn_data.cleaned()))
        <class 'pandas.core.frame.DataFrame'>
        >>> # Population values
        >>> population_data = data_loader.population()
        >>> print(population_data.citation)
        ...
        >>> print(type(population_data.cleaned()))
        <class 'pandas.core.frame.DataFrame'>
        >>> # OxCGRT: Government responses
        >>> oxcgrt_data = data_loader.oxcgrt()
        >>> print(oxcgrt_data.citation)
        ...
        >>> print(type(oxcgrt_data.cleaned()))
        <class 'pandas.core.frame.DataFrame'>
        >>> # Citation list of COVID-19 Data Hub
        >>> print(data_loader.covid19dh_citation)
        ...
    """
    GITHUB_URL = "https://raw.githubusercontent.com"

    def __init__(self, directory="input", update_interval=12):
        # Directory
        try:
            self.dir_path = Path(directory)
        except TypeError:
            raise TypeError(
                f"@directory should be a path-like object, but {directory} was applied.")
        self.update_interval = self.ensure_natural_int(
            update_interval, name="update_interval", include_zero=True)
        # Create the directory if not exist
        self.dir_path.mkdir(parents=True, exist_ok=True)
        # COVID-19 Data Hub
        self.covid19dh = None
        # Cache instances
        self._linelist_data = None

    @staticmethod
    def _last_updated_local(path):
        """
        Return the date last updated of local file/directory.

        Args:
            path (str or pathlibPath): name of the file/directory

        Returns:
            (datetime.datetime): time last updated (UTC)
        """
        m_time = Path(path).stat().st_mtime
        date = datetime.fromtimestamp(m_time)
        return date.astimezone(timezone.utc).replace(tzinfo=None)

    def _download_necessity(self, filename):
        """
        Return whether we need to get the data from remote servers or not,
        comparing the last update of the files.

        Args:
            filename (str): filename of the local file

        Returns:
            (bool): whether we need to get the data from remote servers or not

        Notes:
            If the last updated date is unknown, returns True.
            If @self.update_interval hours have passed and the remote file was updated, return True.
        """
        if not Path(filename).exists():
            return True
        date_local = self._last_updated_local(filename)
        time_limit = date_local + timedelta(hours=self.update_interval)
        return datetime.now() > time_limit

    def _covid19dh(self, name, basename="covid19dh.csv", verbose=True):
        """
        Return the datasets of COVID-19 Data Hub.

        Args:
            name (str): name of dataset, "jhu", "population" or "oxcgrt"
            basename (str): basename of CSV file to save records
            verbose (int): level of verbosity

        Notes:
            If @verbose is 2, detailed citation list will be shown when downloading.
            If @verbose is 1, how to show the list will be explained.
            Citation of COVID-19 Data Hub will be set as JHUData.citation etc.

        Returns:
            covsirphy.CleaningBase: the dataset
        """
        filename, force = self.dir_path.joinpath(basename), False
        if self.covid19dh is None:
            self.covid19dh = COVID19DataHub(filename=filename)
            force = self._download_necessity(filename)
        return self.covid19dh.load(name=name, force=force, verbose=verbose)

    @ property
    def covid19dh_citation(self):
        """
        Return the list of primary sources of COVID-19 Data Hub.
        """
        if self.covid19dh is None:
            self._covid19dh(name="jhu", verbose=0)
        return self.covid19dh.primary

    def jhu(self, basename="covid19dh.csv", local_file=None, verbose=1):
        """
        Load the dataset regarding the number of cases using local CSV file or COVID-19 Data Hub.

        Args:
            basename (str or None): basename of the file to save the data
            local_file (str or None): if not None, load the data from this file
            verbose (int): level of verbosity

        Notes:
            If @verbose is 2, detailed citation list will be shown when downloading.
            If @verbose is 1, how to show the list will be explained.
            Citation of COVID-19 Data Hub will be set as JHUData.citation.

        Returns:
            covsirphy.JHUData: dataset regarding the number of cases
        """
        if local_file is not None:
            return JHUData(filename=local_file)
        # Retrieve JHU data from COVID-19 Data Hub
        jhu_data = self._covid19dh(
            name="jhu", basename=basename, verbose=verbose)
        # Replace Japan dataset with the government-announced data
        japan_data = self.japan()
        jhu_data.replace(japan_data)
        return jhu_data

    def population(self, basename="covid19dh.csv", local_file=None, verbose=1):
        """
        Load the dataset regarding population values using local CSV file or COVID-19 Data Hub.

        Args:
            basename (str or None): basename of the file to save the data
            local_file (str or None): if not None, load the data from this file
            verbose (int): level of verbosity

        Notes:
            If @verbose is 2, detailed citation list will be shown when downloading.
            If @verbose is 1, how to show the list will be explained.
            Citation of COVID-19 Data Hub will be set as PopulationData.citation.

        Returns:
            covsirphy.PopulationData: dataset regarding population values
        """
        if local_file is not None:
            return PopulationData(filename=local_file)
        return self._covid19dh(name="population", basename=basename, verbose=verbose)

    def oxcgrt(self, basename="covid19dh.csv", local_file=None, verbose=1):
        """
        Load the dataset regarding OxCGRT data using local CSV file or COVID-19 Data Hub.

        Args:
            basename (str or None): basename of the file to save the data
            local_file (str or None): if not None, load the data from this file
            verbose (int): level of verbosity

        Notes:
            If @verbose is 2, detailed citation list will be shown when downloading.
            If @verbose is 1, how to show the list will be explained.
            Citation of COVID-19 Data Hub will be set as OxCGRTData.citation.

        Returns:
            covsirphy.JHUData: dataset regarding OxCGRT data
        """
        if local_file is not None:
            return OxCGRTData(filename=local_file)
        return self._covid19dh(name="oxcgrt", basename=basename, verbose=verbose)

    def japan(self, basename="covid_japan.csv", local_file=None, verbose=1):
        """
        Load the dataset of the number of cases in Japan.
        https://github.com/lisphilar/covid19-sir/tree/master/data

        Args:
            basename (str): basename of the file to save the data
            local_file (str or None): if not None, load the data from this file
            verbose (int): level of verbosity

        Returns:
            covsirphy.CountryData: dataset at country level in Japan
        """
        filename = local_file or self.dir_path.joinpath(basename)
        force = self._download_necessity(filename=filename)
        return JapanData(filename=filename, force=force, verbose=verbose)

    def linelist(self, basename="linelist.csv", verbose=1):
        """
        Load linelist of case reports.
        https://github.com/beoutbreakprepared/nCoV2019

        Args:
            basename (str): basename of the file to save the data
            verbose (int): level of verbosity

        Returns:
            covsirphy.CountryData: dataset at country level in Japan
        """
        if self._linelist_data is None:
            filename = self.dir_path.joinpath(basename)
            force = self._download_necessity(filename=filename)
            self._linelist_data = LinelistData(
                filename=filename, force=force, verbose=verbose)
        return self._linelist_data

    def pcr(self, basename="covid19dh.csv", local_file=None,
            basename_owid="ourworldindata_pcr.csv", verbose=1):
        """
        Load the dataset regarding the number of tests and confirmed cases,
        using local CSV file or COVID-19 Data Hub.

        Args:
            basename (str or None): basename of the file to save "COVID-19 Data Hub" data
            local_file (str or None): if not None, load the data from this file
            basename_owid (str): basename of the file to save "Our World In Data" data
            verbose (int): level of verbosity

        Notes:
            If @verbose is 2, detailed citation list will be shown when downloading.
            If @verbose is 1, how to show the list will be explained.
            Citation of COVID-19 Data Hub will be set as JHUData.citation.

        Returns:
            covsirphy.PCRData: dataset regarding the number of tests and confirmed cases
        """
        if local_file is not None:
            return PCRData(filename=local_file)
        # Retrieve JHU data from COVID-19 Data Hub
        pcr_data = self._covid19dh(
            name="pcr", basename=basename, verbose=verbose)
        # Update the values using "Our World In Data" dataset
        owid_filename = self.dir_path.joinpath(basename_owid)
        owid_force = self._download_necessity(filename=owid_filename)
        pcr_data.use_ourworldindata(filename=owid_filename, force=owid_force)
        # Replace Japan dataset with the government-announced data
        japan_data = self.japan()
        pcr_data.replace(japan_data)
        return pcr_data

    def vaccine(self, basename="ourworldindata_vaccine.csv", verbose=1):
        """
        Load the dataset regarding vaccination.
        https://github.com/owid/covid-19-data/tree/master/public/data
        https://ourworldindata.org/coronavirus

        Args:
            basename (str): basename of the file to save the data
            verbose (int): level of verbosity

        Returns:
            covsirphy.VaccineData: dataset regarding vaccines
        """
        filename = self.dir_path.joinpath(basename)
        force = self._download_necessity(filename=filename)
        return VaccineData(filename=filename, force=force, verbose=verbose)

    def pyramid(self, basename="wbdata_population_pyramid.csv", verbose=1):
        """
        Load the dataset regarding population pyramid.
        World Bank Group (2020), World Bank Open Data, https://data.worldbank.org/

        Args:
            basename (str): basename of the file to save the data
            verbose (int): level of verbosity

        Returns:
            covsirphy.PopulationPyramidData: dataset regarding population pyramid
        """
        filename = self.dir_path.joinpath(basename)
        return PopulationPyramidData(filename=filename, force=False, verbose=verbose)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import warnings


def deprecate(old, new=None, version=None):
    """
    Decorator to raise deprecation warning.

    Args:
        old (str): description of the old method/function
        new (str or None): description of the new method/function
        version (str or None): version number, like 2.7.3-alpha
    """
    def _deprecate(func):
        def wrapper(*args, **kwargs):
            if new is None:
                comment = f"{old} is deprecated, version > {version}"
            else:
                comment = f"Please use {new} rather than {old}, version > {version}"
            warnings.warn(
                comment,
                DeprecationWarning,
                stacklevel=2
            )
            return func(*args, **kwargs)
        return wrapper
    return _deprecate


class SubsetNotFoundError(KeyError, ValueError):
    """
    Error when subset was failed with specified arguments.

    Args:
        country (str): country name
        country_alias (str or None): country name used in the dataset
        province (str or None): province name
        start_date (str or None): start date, like 22Jan2020
        end_date (str or None): end date, like 01Feb2020
        date (str or None): specified date, like 22Jan2020
        message (str or None): the other messages
    """

    def __init__(self, country, country_alias=None, province=None,
                 start_date=None, end_date=None, date=None, message=None):
        self.area = self._area(country, country_alias, province)
        self.date = self._date(start_date, end_date, date)
        self.message = "" if message is None else f" {message}"

    def __str__(self):
        return f"Records{self.message} in {self.area}{self.date} were not found."

    @staticmethod
    def _area(country, country_alias, province):
        """
        Error when subset was failed with specified arguments.

        Args:
            country (str): country name
            country_alias (str or None): country name used in the dataset
            province (str or None): province name

        Returns:
            str: area name
        """
        if country_alias is None or country == country_alias:
            c_alias_str = ""
        else:
            c_alias_str = f" ({country_alias})"
        province_str = "" if province is None else f"{province}, "
        return f"{province_str}{country}{c_alias_str}"

    @staticmethod
    def _date(start_date, end_date, date):
        """
        Error when subset was failed with specified arguments.

        Args:
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020
            date (str or None): specified date, like 22Jan2020
        """
        if date is not None:
            return f" on {date}"
        start_str = "" if start_date is None else f" from {start_date}"
        end_str = "" if end_date is None else f" to {end_date}"
        return f"{start_str}{end_str}"


class ScenarioNotFoundError(KeyError):
    """
    Error when unregistered scenario name was specified.

    Args:
        name (str): scenario name
    """

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"{self.name} scenario is not registered."


class UnExecutedError(AttributeError, NameError, ValueError):
    """
    Error when we have unexecuted methods that we need to run in advance.

    Args:
        method_name (str): method name to run in advance
        message (str or None): the other messages
    """

    def __init__(self, method_name, message=None):
        self.method_name = method_name
        self.message = "." if message is None else f" {message}."

    def __str__(self):
        return f"Please execute {self.method_name} in advance{self.message}"


class PCRIncorrectPreconditionError(KeyError):
    """
    Error when checking preconditions in the PCR data.

    Args:
        country (str): country name
        province (str or None): province name
        message (str or None): the other messages
    """

    def __init__(self, country, province=None, message=None):
        self.area = self._area(country, province)
        self.message = "" if message is None else f" {message}"

    def __str__(self):
        return f"{self.message}{self.area}."

    @staticmethod
    def _area(country, province):
        """
        Error when PCR preconditions failed with specified arguments.

        Args:
            country (str): country name
            province (str or None): province name

        Returns:
            str: area name
        """
        if province == "-":
            province = None
        country_str = (" in country " + country) if not province else ""
        province_str = "" if province is None else (" in province " + province)
        return f"{province_str}{country_str}"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
from covsirphy.cleaning.term import Term
from covsirphy.phase.phase_unit import PhaseUnit
from covsirphy.phase.sr_change import ChangeFinder


class PhaseSeries(Term):
    """
    A series of phases.

    Args:
        first_date (str): the first date of the series, like 22Jan2020
        last_date (str): the last date of the records, like 25May2020
        population (int): initial value of total population in the place
    """

    def __init__(self, first_date, last_date, population):
        self.first_date = self.ensure_date(first_date, "first_date")
        self.last_date = self.ensure_date(last_date, "last_date")
        self.init_population = self.ensure_population(population)
        # List of PhaseUnit
        self._units = []
        self.clear(include_past=True)

    def __iter__(self):
        yield from self._units

    def __len__(self):
        return len([unit for unit in self._units if unit])

    def unit(self, phase="last"):
        """
        Return the unit of the phase.

        Args:
            phase (str): phase name (1st etc.) or "last"

        Returns:
            covsirphy.PhaseUnit: the unit of the phase

        Notes:
            When @phase is 'last' and no phases were registered, returns A phase
            with the start/end dates are the previous date of the first date and initial population value.
        """
        if phase == "last":
            if self._units:
                return self._units[-1]
            pre_date = self.yesterday(self.first_date)
            return PhaseUnit(pre_date, pre_date, self.init_population)
        num = self.str2num(phase)
        try:
            return self._units[num]
        except IndexError:
            raise KeyError(f"{phase} phase is not registered.")

    def clear(self, include_past=False):
        """
        Clear phase information. Future phases will be always deleted.

        Args:
            include_past (bool): if True, include past phases.

        Returns:
            covsirphy.PhaseSeries: self
        """
        if include_past:
            self._units = []
        self._units = [unit for unit in self._units if unit <= self.last_date]
        return self

    def _calc_end_date(self, start_date, end_date=None, days=None):
        """
        Return the end date.

        Args:
            start_date (str): start date of the phase
            end_date (str): end date of the past phase, like 22Jan2020
            days (int or None): the number of days to add

        Returns:
            str: end date
        """
        if end_date is not None:
            self.ensure_date_order(start_date, end_date, name="end_date")
            return end_date
        if days is None:
            return self.last_date
        return self.date_change(start_date, days=days - 1)

    def add(self, end_date=None, days=None, population=None, model=None, tau=None, **kwargs):
        """
        Add a past phase.

        Args:
            end_date (str): end date of the past phase, like 22Jan2020
            days (int or None): the number of days to add
            population (int or None): population value
            model (covsirphy.ModelBase): ODE model
            tau (int or None): tau value [min], a divisor of 1440 (prioritize the previous value)
            kwargs: keyword arguments of model parameters

        Returns:
            covsirphy.PhaseSeries: self

        Notes:
            If @population is None, the previous initial value will be used.
            When addition of past phases was not completed and the new phase is future phase, fill in the blank.
        """
        last_unit = self.unit(phase="last")
        # Basic information
        start_date = self.tomorrow(last_unit.end_date)
        end_date = self._calc_end_date(
            start_date, end_date=end_date, days=days)
        population = self.ensure_population(population or last_unit.population)
        model = model or last_unit.model
        tau = last_unit.tau or tau
        if model is None:
            param_dict = {}
        else:
            param_dict = {
                k: v for (k, v) in {**last_unit.to_dict(), **kwargs}.items()
                if k in model.PARAMETERS}
        # Create PhaseUnit
        unit = PhaseUnit(start_date, end_date, population)
        # Add phase if the last date is not included
        if self.last_date not in unit or unit <= self.last_date:
            unit.set_ode(model=model, tau=tau, **param_dict)
            self._units.append(unit)
            return self
        # Fill in the blank of past dates
        filling = PhaseUnit(start_date, self.last_date, population)
        filling.set_ode(model=model, tau=tau, **param_dict)
        target = PhaseUnit(
            self.tomorrow(self.last_date), end_date, population)
        target.set_ode(model=model, tau=tau, **param_dict)
        # Add new phase
        self._units.extend([filling, target])
        return self

    def delete(self, phase="last"):
        """
        Delete a phase. The phase will be combined to the previous phase.

        Args:
            phase (str): phase name, like 0th, 1st, 2nd... or 'last'

        Returns:
            covsirphy.PhaseSeries: self

        Notes:
            When @phase is '0th', disable 0th phase. 0th phase will not be deleted.
            When @phase is 'last', the last phase will be deleted.
        """
        if phase == "0th":
            self.disable("0th")
            return self
        if self.unit(phase) == self.unit("last"):
            self._units = self._units[:-1]
            return self
        phase_pre = self.num2str(self.str2num(phase) - 1)
        unit_pre, unit_fol = self.unit(phase_pre), self.unit(phase)
        if unit_pre <= self.last_date and unit_fol >= self.last_date:
            phase_next = self.num2str(self.str2num(phase) + 1)
            unit_next = self.unit(phase_next)
            model = unit_next.model
            param_dict = {
                k: v for (k, v) in unit_next.to_dict().items() if k in model.PARAMETERS}
            unit_new = PhaseUnit(
                unit_fol.start_date, unit_next.end_date, unit_next.population)
            unit_new.set_ode(model=model, tau=unit_next.tau, **param_dict)
            return self
        unit_new = PhaseUnit(
            unit_pre.start_date, unit_fol.end_date, unit_pre.population)
        model = unit_pre.model
        if model is None:
            param_dict = {}
        else:
            param_dict = {
                k: v for (k, v) in unit_pre.to_dict().items() if k in model.PARAMETERS}
        unit_new.set_ode(model=model, tau=unit_pre.tau, **param_dict)
        units = [
            unit for unit in [unit_new, *self._units] if unit not in [unit_pre, unit_fol]]
        self._units = sorted(units)
        return self

    def disable(self, phase):
        """
        The phase will be disabled and removed from summary.

        Args:
            phase (str): phase name, like 0th, 1st, 2nd...

        Returns:
            covsirphy.PhaseSeries: self
        """
        phase_id = self.str2num(phase)
        self._units[phase_id].disable()
        return self

    def enable(self, phase):
        """
        The phase will be enabled and appears in summary.

        Args:
            phase (str): phase name, like 0th, 1st, 2nd...

        Returns:
            covsirphy.PhaseSeries: self
        """
        phase_id = self.str2num(phase)
        self._units[phase_id].enable()
        return self

    def summary(self):
        """
        Summarize the series of phases in a dataframe.

        Returns:
            (pandas.DataFrame):
                Index:
                    - phase name, like 1st, 2nd, 3rd...
                Columns:
                    - Type: 'Past' or 'Future'
                    - Start: start date of the phase
                    - End: end date of the phase
                    - Population: population value of the start date
                    - other information registered to the phases
        """
        info_dict = self.to_dict()
        if not info_dict:
            return pd.DataFrame(columns=[self.TENSE, self.START, self.END, self.N])
        # Convert to dataframe
        df = pd.DataFrame.from_dict(info_dict, orient="index")
        return df.dropna(how="all", axis=1).fillna(self.UNKNOWN)

    def to_dict(self):
        """
        Summarize the series of phase in a dictionary.

        Returns:
            (dict): nested dictionary of phase information
                - key (str): phase number, like 1th, 2nd,...
                - value (dict): phase information
                    - 'Type': (str) 'Past' or 'Future'
                    - values of PhaseUnit.to_dict()
        """
        return {
            self.num2str(phase_id): {
                self.TENSE: self.PAST if unit <= self.last_date else self.FUTURE,
                **unit.to_dict()
            }
            for (phase_id, unit) in enumerate(self._units) if unit
        }

    def replace(self, phase, new):
        """
        Replace phase object.

        Args:
            phase (str): phase name, like 0th, 1st, 2nd...
            new (covsirphy.PhaseUnit): new phase object

        Returns:
            covsirphy.PhaseSeries: self
        """
        old = self.unit(phase)
        if old != new:
            raise ValueError(
                "Combination of start/end date is different. old: {old}, new: {new}")
        units = [unit for unit in self._units if unit != old]
        self._units = sorted(units + [new])
        return self

    def replaces(self, phase=None, new_list=None, keep_old=False):
        """
        Replace phase object.

        Args:
            phase (str or None): phase name, like 0th, 1st, 2nd...
            new_list (list[covsirphy.PhaseUnit]): new phase objects

        Returns:
            covsirphy.PhaseSeries: self

        Notes:
            If @phase is None and @keep_old is False, all old phases will be deleted.
            If @phase is not None, the phase will be deleted.
            @new_list must be specified.
        """
        # Check arguments
        if not isinstance(new_list, list):
            raise TypeError("@new_list must be a list of covsirphy.PhaseUnit.")
        type_ok = all(isinstance(unit, PhaseUnit) for unit in new_list)
        if not type_ok:
            raise TypeError("@new_list must be a list of covsirphy.PhaseUnit.")
        # Replace phases
        if phase is None:
            old_units = [
                unit for unit in self._units if unit not in new_list] if keep_old else []
        else:
            exc_unit = self.unit(phase)
            old_units = [unit for unit in self._units if unit != exc_unit]
        self._units = self._ensure_series(old_units + new_list)

    @classmethod
    def _ensure_series(cls, units):
        """
        Ensure that the list is a series of phases.

        Args:
            units (list[covsirphy.PhaseUnit]): list of units

        Returns:
            list[covsirphy.PhaseUnit]: sorted list of units

        Raises:
            ValueError: Phases are not series.
        """
        sorted_units = sorted(units)
        s = ", ".join([str(unit) for unit in sorted_units])
        for (i, unit) in enumerate(sorted_units):
            if i in [0, len(sorted_units) - 1]:
                continue
            sta_app = unit.start_date
            end_app = unit.end_date
            sta = cls.tomorrow(sorted_units[i - 1].end_date)
            end = cls.yesterday(sorted_units[i + 1].start_date)
            if sta != sta_app or end != end_app:
                raise ValueError(
                    f"The list of units does not a series of phases. Applied: {s}")
        return sorted_units

    def trend(self, sr_df, **kwargs):
        """
        Perform S-R trend analysis.

        Args:
            sr_df (pandas.DataFrame): susceptible and recovered records
                Index:
                    Date (pd.TimeStamp): Observation date
                Columns:
                    - Recovered (int): the number of recovered cases (> 0)
                    - Susceptible (int): the number of susceptible cases
                    - any other columns will be ignored
            kwargs: keyword arguments of ChangeFinder()

        Returns:
            covsirphy.PhaseSeries: self
        """
        sta = self.date_obj(self.first_date)
        end = self.date_obj(self.last_date)
        sr_df = sr_df.loc[(sr_df.index >= sta) & (sr_df.index <= end), :]
        # Find change points
        finder = ChangeFinder(sr_df, **kwargs)
        finder.run()
        # Register phases
        self.clear(include_past=True)
        _, end_dates = finder.date_range()
        [self.add(end_date=end_date) for end_date in end_dates]
        return self

    def trend_show(self, sr_df, area=None, filename=None):
        """
        Show S-R plane, indicating change points found with S-R trend analysis.

        Args:
            sr_df (pandas.DataFrame): susceptible and recovered records
                Index:
                    Date (pd.TimeStamp): Observation date
                Columns:
                    - Recovered (int): the number of recovered cases (> 0)
                    - Susceptible (int): the number of susceptible cases
                    - any other columns will be ignored
            area (str or None): area name
            filename (str): filename of the figure, or None (display)
            kwargs: keyword arguments of ChangeFinder()

        Returns:
            covsirphy.PhaseSeries: self
        """
        area = area or self.UNKNOWN
        sta = self.date_obj(self.first_date)
        end = self.date_obj(self.last_date)
        sr_df = sr_df.loc[(sr_df.index >= sta) & (sr_df.index <= end), :]
        finder = ChangeFinder(sr_df)
        change_dates = [
            unit.start_date for unit in self._units[1:] if unit <= self.last_date]
        finder.show(
            area=area, change_dates=change_dates, filename=filename)

    def simulate(self, record_df, y0_dict=None):
        """
        Simulate ODE models with set parameter values.

        Args:
            record_df (pandas.DataFrame):
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases (> 0)
                    - Susceptible (int): the number of susceptible cases
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value

        Returns:
            pandas.DataFrame
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Country (str): country/region name
                    - Province (str): province/prefecture/state name
                    - Variables of the model and dataset (int): Confirmed etc.
        """
        dataframes = []
        rec_dates = record_df[self.DATE].dt.strftime(self.DATE_FORMAT).unique()
        for (num, unit) in enumerate(self._units):
            if not unit:
                continue
            if unit.start_date in rec_dates:
                unit.set_y0(record_df)
            else:
                try:
                    unit.set_y0(dataframes[-1])
                except IndexError:
                    pass
            df = unit.simulate(y0_dict=y0_dict)
            dataframes.append(df)
        sim_df = pd.concat(dataframes, ignore_index=True, sort=True)
        sim_df = sim_df.set_index(self.DATE).resample("D").last()
        sim_df = sim_df.dropna().astype(np.int64)
        return sim_df.reset_index()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
try:
    import better_exceptions
    better_exceptions_installed = True
except ImportError:
    better_exceptions_installed = False
# version
from covsirphy.__version__ import __version__
# util
from covsirphy.util.plotting import line_plot
from covsirphy.util.map import jpn_map
from covsirphy.util.optimize import Optimizer
from covsirphy.util.stopwatch import StopWatch
from covsirphy.util.error import deprecate
from covsirphy.util.error import SubsetNotFoundError, ScenarioNotFoundError, UnExecutedError
from covsirphy.util.error import PCRIncorrectPreconditionError
from covsirphy.util.file import save_dataframe
from covsirphy.util.argument import find_args
# cleaning
from covsirphy.cleaning.term import Term, Word
from covsirphy.cleaning.cbase import CleaningBase
from covsirphy.cleaning.jhu_data import JHUData
from covsirphy.cleaning.jhu_complement import JHUDataComplementHandler
from covsirphy.cleaning.country_data import CountryData
from covsirphy.cleaning.japan_data import JapanData
from covsirphy.cleaning.population import PopulationData, Population
from covsirphy.cleaning.pyramid import PopulationPyramidData
from covsirphy.cleaning.oxcgrt import OxCGRTData
from covsirphy.cleaning.pcr_data import PCRData
from covsirphy.cleaning.covid19datahub import COVID19DataHub
from covsirphy.cleaning.linelist import LinelistData
from covsirphy.cleaning.vaccine_data import VaccineData
from covsirphy.cleaning.dataloader import DataLoader
# ode
from covsirphy.ode.mbase import ModelBase
from covsirphy.ode.sir import SIR
from covsirphy.ode.sird import SIRD
from covsirphy.ode.sirf import SIRF
from covsirphy.ode.sirfv import SIRFV
from covsirphy.ode.sewirf import SEWIRF
# simulation
from covsirphy.simulation.estimator import Estimator
from covsirphy.simulation.simulator import ODESimulator
# phase
from covsirphy.phase.trend import Trend
from covsirphy.phase.sr_change import ChangeFinder
from covsirphy.phase.phase_unit import PhaseUnit
from covsirphy.phase.phase_series import PhaseSeries
from covsirphy.phase.phase_estimator import MPEstimator
# analysis
from covsirphy.analysis.example_data import ExampleData
from covsirphy.analysis.data_handler import DataHandler
from covsirphy.analysis.param_tracker import ParamTracker
from covsirphy.analysis.scenario import Scenario
from covsirphy.analysis.model_validator import ModelValidator
# worldwide
from covsirphy.worldwide.policy import PolicyMeasures


def get_version():
    """
    Return the version number, like CovsirPhy v0.0.0
    """
    return f"CovsirPhy v{__version__}"


__all__ = [
    "ExampleData", "Scenario", "ModelValidator", "ParamTracker",
    "ODESimulator", "ChangeFinder", "DataHandler",
    "PhaseSeries", "PhaseUnit", "MPEstimator",
    "Term", "CleaningBase", "DataLoader", "COVID19DataHub",
    "JHUData", "CountryData", "PopulationData", "OxCGRTData", "VaccineData",
    "PopulationPyramidData",
    "LinelistData", "PCRData", "JapanData", "JHUDataComplementHandler",
    "ModelBase", "SIR", "SIRD", "SIRF", "SIRFV", "SEWIRF",
    "Estimator", "Trend", "Optimizer",
    "line_plot", "jpn_map", "StopWatch", "deprecate", "find_args",
    "save_dataframe",
    "PolicyMeasures",
    "SubsetNotFoundError", "ScenarioNotFoundError", "UnExecutedError",
    "PCRIncorrectPreconditionError",
    # Deprecated
    "Population", "Word",
]

# Show exceptions in better format if used from command line
if not hasattr(sys, "ps1") or not better_exceptions_installed:
    better_exceptions.MAX_LENGTH = None
    better_exceptions.hook()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import functools
import sys
import warnings
import matplotlib
if not hasattr(sys, "ps1"):
    matplotlib.use("Agg")
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.optimize import curve_fit, OptimizeWarning
from covsirphy.cleaning.term import Term


class Trend(Term):
    """
    S-R trend analysis in a phase.

    Args:
        sr_df (pandas.DataFrame)
            Index:
                Date (pd.TimeStamp): Observation date
            Columns:
                - Recovered (int): the number of recovered cases (> 0)
                - Susceptible (int): the number of susceptible cases
                - any other columns will be ignored
    """
    L = "linear"
    N = "negative_exponential"

    def __init__(self, sr_df):
        # Dataset
        self.sr_df = self.ensure_dataframe(
            sr_df, name="sr_df", time_index=True, columns=[self.S, self.R])
        # Dataset for analysis
        if len(self.sr_df) < 3:
            raise ValueError("The length of @sr_df must be over 2.")
        # Setting for analysis
        self._result_df = pd.DataFrame()
        self.fit_fnc = self.linear

    @property
    def result_df(self):
        """
        pandas.DataFrame: results of fitting
            Index:
                - index (Date) (pd.TimeStamp): Observation date
            Columns:
                - Recovered: The number of recovered cases
                - Susceptible_actual: Actual values of Susceptible
                - columns defined by @columns
        """
        if self._result_df.empty:
            self.run()
        return self._result_df

    def run(self):
        """
        Perform curve fitting with some functions and select the best solution.
        Then, the result and RMSLE score of the best solution will be saved.

        Returns:
            covsirphy.Trend: self
        """
        L, N = self.L, self.N
        # Perform fitting and calculate RMSLE scores
        dataframe_dict = {func: self._run(func=func) for func in (L, N)}
        score_dict = {
            func: self._rmsle(fit_df) for (func, fit_df) in dataframe_dict.items()}
        # Select the best dataframe
        if 0 < score_dict[L] < score_dict[N] or not score_dict[N]:
            self._result_df = dataframe_dict[L]
        else:
            self._result_df = dataframe_dict[N]
        return self

    def _run(self, func):
        """
        Perform curve fitting of S-R trend with linear or negative exponential function and save the result.

        Args:
            func (function): the selected curve fitting function, either linear or negative exponential

        Returns:
            pandas.DataFrame: results of fitting
                Index:
                    - index (Date) (pd.TimeStamp): Observation date
                Columns:
                    - Recovered: The number of recovered cases
                    - Susceptible_actual: Actual values of Susceptible
                    - columns defined by @columns
        """
        self.fit_fnc = self.negative_exp if func == self.N else self.linear
        return self._fitting(self.sr_df)

    def _fitting(self, sr_df):
        """
        Perform curve fitting of S-R trend with linear or negative exponential function.

        Args:
            sr_df (pandas.DataFrame): training dataset
                Index:
                    - index (Date) (pd.TimeStamp): Observation date
                Columns:
                    - Recovered: The number of recovered cases
                    - Susceptible: Actual data of Susceptible

        Returns:
            pandas.DataFrame
                Index:
                    - index (Date) (pd.TimeStamp): Observation date
                Columns:
                    - Recovered (int): The number of recovered cases
                    - Susceptible_actual (int): Actual values of Susceptible
                    - Susceptible_predicted (int): Predicted values of Susceptible
        """
        df = sr_df.rename({self.S: f"{self.S}{self.A}"}, axis=1)
        df = df.astype(np.float64)
        # Calculate initial values of parameters
        x_series = df[self.R]
        y_series = np.log(df[f"{self.S}{self.A}"]).astype(np.float64)
        a_ini = y_series.max()
        b_ini = y_series.diff().reset_index(drop=True)[1] / a_ini
        # Curve fitting with linear or negative exponential function
        warnings.simplefilter("ignore", OptimizeWarning)
        warnings.simplefilter("ignore", RuntimeWarning)
        param, _ = curve_fit(
            self.fit_fnc, x_series, y_series,
            p0=[a_ini, b_ini],
            # Increase mux number of iteration in curve fitting from 600 (default)
            maxfev=5000
        )
        # Predict the values with the parameters
        f_partial = functools.partial(
            self.fit_fnc, a=param[0], b=param[1]
        )
        df[f"{self.S}{self.P}"] = np.exp(
            f_partial(x_series)).astype(np.float64)
        return df.astype(np.int64, errors="ignore")

    def rmsle(self):
        """
        Return the best RMSLE score.

        Returns:
            (float): RMSLE score
        """
        if self._result_df.empty:
            self.run()
        return self._rmsle(self._result_df)

    def _rmsle(self, fit_df):
        """
        Calculate RMSLE score of actual/predicted Susceptible.

        Args:
            fit_df (pandas.DataFrame):
                Index:
                    - index (Date) (pd.TimeStamp): Observation date
                Columns:
                    - Recovered (int): The number of recovered cases
                    - Susceptible_actual (int): Actual values of Susceptible
                    - Susceptible_predicted (int): Predicted values of Susceptible

        Returns:
            (float): RMSLE score
        """
        df = fit_df.replace(np.inf, 0)
        df = df.loc[df[f"{self.S}{self.A}"] > 0, :]
        df = df.loc[df[f"{self.S}{self.P}"] > 0, :]
        actual = df[f"{self.S}{self.A}"]
        predicted = df[f"{self.S}{self.P}"]
        # Calculate RMSLE score
        scores = np.abs(
            np.log10(actual + 1) - np.log10(predicted + 1)
        )
        return scores.sum()

    def show(self, area, filename=None):
        """
        show the result as a figure.

        Args:
            area (str): area name
            filename (str): filename of the figure, or None (display)
        """
        df = self._result_df.copy()
        df = df.rename({f"{self.S}{self.P}": "Predicted"}, axis=1)
        # Star/end date
        start_date = self.sr_df.index.min().strftime(self.DATE_FORMAT)
        end_date = self.sr_df.index.max().strftime(self.DATE_FORMAT)
        # Plotting
        title = f"{area}: S-R trend from {start_date} to {end_date}"
        self.show_with_many(
            result_df=df,
            predicted_cols=["Predicted"],
            title=title,
            filename=filename
        )

    @classmethod
    def show_with_many(cls, result_df, predicted_cols,
                       title, vlines=None, filename=None):
        """
        show the result as a figure.

        Args:
            result_df (pandas.DataFrame): results of fitting
                Index:
                    - index (Date) (pd.TimeStamp): Observation date
                Columns:
                    - Recovered: The number of recovered cases
                    - Susceptible_actual: Actual values of Susceptible
                    - columns defined by @predicted_cols
            predicted_cols (list[str]): list of columns which have predicted values
            title (str): title of the figure
            vlines (list[int]): list of Recovered values to show vertical lines
            filename (str): filename of the figure, or None (show figure)
        """
        result_df = cls.ensure_dataframe(
            result_df, name="result_df", time_index=True,
            columns=[cls.R, f"{cls.S}{cls.A}", *predicted_cols]
        )
        df = result_df.replace(np.inf, np.nan)
        x_series = df[cls.R]
        actual = df[f"{cls.S}{cls.A}"]
        # Plot the actual values
        warnings.simplefilter("ignore", FutureWarning)
        plt.plot(
            x_series, actual,
            label="Actual", color="black",
            marker=".", markeredgewidth=0, linewidth=0
        )
        # Plot the predicted values
        for col in predicted_cols:
            plt.plot(x_series, df[col], label=col.replace(cls.P, str()))
        # x-axis
        plt.xlabel(cls.R)
        plt.xlim(0, None)
        # y-axis
        plt.ylabel(cls.S)
        try:
            plt.yscale("log", base=10)
        except Exception:
            plt.yscale("log", basey=10)
        # Delete y-labels of log-scale (minor) axis
        plt.setp(plt.gca().get_yticklabels(minor=True), visible=False)
        plt.gca().tick_params(left=False, which="minor")
        # Set new y-labels of major axis
        ymin, ymax = plt.ylim()
        ydiff_scale = int(np.log10(ymax - ymin))
        yticks = np.linspace(
            round(ymin, - ydiff_scale),
            round(ymax, - ydiff_scale),
            5,
            dtype=np.int64
        )
        plt.gca().set_yticks(yticks)
        fmt = matplotlib.ticker.ScalarFormatter(useOffset=False)
        fmt.set_scientific(False)
        plt.gca().yaxis.set_major_formatter(fmt)
        # Title
        plt.title(title)
        # Vertical lines
        if isinstance(vlines, (list, tuple)):
            for value in vlines:
                plt.axvline(x=value, color="black", linestyle=":")
        # Legend
        plt.legend(
            bbox_to_anchor=(1.02, 0), loc="lower left", borderaxespad=0
        )
        # Save figure or show figure
        warnings.simplefilter("ignore", UserWarning)
        plt.tight_layout()
        if filename is None:
            plt.show()
            return None
        plt.savefig(
            filename, bbox_inches="tight", transparent=False, dpi=300
        )
        plt.clf()
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime
from dask import dataframe as dd
import numpy as np
import pandas as pd
from covsirphy.util.error import deprecate, SubsetNotFoundError
from covsirphy.cleaning.cbase import CleaningBase


class PopulationData(CleaningBase):
    """
    Data cleaning of total population dataset.

    Args:
        filename (str): CSV filename of the dataset
        citation (str): citation
    """
    POPULATION_COLS = [
        CleaningBase.ISO3,
        CleaningBase.COUNTRY,
        CleaningBase.PROVINCE,
        CleaningBase.DATE,
        CleaningBase.N
    ]

    def __init__(self, filename=None, citation=None):
        self.created_time = datetime.now()
        if filename is None:
            self._raw = pd.DataFrame()
            self._cleaned_df = pd.DataFrame(columns=self.POPULATION_COLS)
        else:
            self._raw = dd.read_csv(
                filename, dtype={"Province/State": "object"}
            ).compute()
            self._cleaned_df = self._cleaning()
        self._citation = citation or ""

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.

        Returns:
            pandas.DataFrame
                Index:
                    reset index
                Columns:
                    - ISO3 (str): ISO3 code or "-"
                    - Country (pandas.Category): country/region name
                    - Province (pandas.Category): province/prefecture/state name
                    - Date (pd.TimeStamp): date of the records (if available) or today
                    - Population (int): total population
        """
        df = self._raw.copy()
        # Rename the columns
        df = df.rename(
            {
                "Country.Region": self.COUNTRY,
                "Country/Region": self.COUNTRY,
                "Province.State": self.PROVINCE,
                "Province/State": self.PROVINCE,
                "ObservationDate": self.DATE
            },
            axis=1
        )
        # Confirm the expected columns are in raw data
        expected_cols = [
            self.COUNTRY, self.PROVINCE, self.N
        ]
        self.ensure_dataframe(df, name="the raw data", columns=expected_cols)
        # ISO3
        df[self.ISO3] = df[self.ISO3] if self.ISO3 in df.columns else self.UNKNOWN
        # Date
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        # Country
        df[self.COUNTRY] = df[self.COUNTRY].replace(
            {
                # COD
                "Congo, the Democratic Republic of the": "Democratic Republic of the Congo",
                # COG
                "Congo": "Republic of the Congo",
            }
        )
        # Province
        df[self.PROVINCE] = df[self.PROVINCE].fillna(self.UNKNOWN)
        df.loc[df[self.COUNTRY] == "Diamond Princess", [
            self.COUNTRY, self.PROVINCE]] = ["Others", "Diamond Princess"]
        # Values
        df = df.dropna(subset=[self.N]).reset_index(drop=True)
        df[self.N] = df[self.N].astype(np.int64)
        # Columns to use
        df = df.loc[
            :, [self.ISO3, self.COUNTRY, self.PROVINCE, self.DATE, self.N]]
        # Remove duplicates
        df = df.drop_duplicates().reset_index(drop=True)
        # Update data types to reduce memory
        df[self.AREA_ABBR_COLS] = df[self.AREA_ABBR_COLS].astype("category")
        return df

    def _created_date(self):
        """
        Return 0:00 AM of the created date.

        Returns:
            datetime.datetime
        """
        return self.created_time.replace(hour=0, minute=0, second=0, microsecond=0)

    def total(self):
        """
        Return the total value of population in the dataset.

        Returns:
            int
        """
        values = self._cleaned_df[self.N]
        return int(sum(values))

    def to_dict(self, country_level=True):
        """
        Return dictionary of population values.

        Args:
        country_level (str): whether key is country name or not

        Returns:
            dict
                - if @country_level is True, {"country", population}
                - if False, {"country/province", population}
        """
        df = self._cleaned_df.copy()
        if country_level:
            df = df.loc[df[self.PROVINCE] == self.UNKNOWN, :]
            df["key"] = df[self.COUNTRY]
        else:
            df = df.loc[df[self.PROVINCE] != self.UNKNOWN, :]
            df["key"] = df[self.COUNTRY].str.cat(
                df[self.PROVINCE], sep=self.SEP
            )
        return df.set_index("key").to_dict()[self.N]

    def value(self, country, province=None, date=None):
        """
        Return the value of population in the place.

        Args:
            country (str): country name or ISO3 code
            province (str): province name
            date (str or None): observation date, like 01Jun2020

        Returns:
            int: population in the place

        Notes:
            If @date is None, the created date of the instancewill be used
        """
        country_alias = self.ensure_country_name(country)
        try:
            df = self.subset(
                country=country, province=province, start_date=date, end_date=date)
        except KeyError:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=province, date=date)
        df = df.sort_values(self.DATE)
        return int(df.loc[df.index[-1], [self.N]].values[0])

    def update(self, value, country, province=None, date=None):
        """
        Update the value of a new place.

        Args:
            value (int): population in the place
            country (str): country name
            province (str): province name
            date (str or None): observation date, like 01Jun2020

        Returns:
            covsirphy.PopulationData: self

        Notes:
            If @date is None, the created date of the instance will be used.
            If @province is None, "-" will be used.
        """
        population = self.ensure_natural_int(value, "value")
        province = province or self.UNKNOWN
        date_stamp = pd.to_datetime(date or self._created_date())
        df = self._cleaned_df.copy()
        c_series = df[self.COUNTRY]
        p_series = df[self.PROVINCE]
        d_series = df[self.DATE]
        sel = (c_series == country) & (
            p_series == province) & (d_series == date_stamp)
        if not df.loc[sel, :].empty:
            df.loc[sel, self.N] = value
            self._cleaned_df = df.copy()
            return self
        series = pd.Series(
            [self.UNKNOWN, country, province, date_stamp, population],
            index=[self.ISO3, self.COUNTRY, self.PROVINCE, self.DATE, self.N]
        )
        self._cleaned_df = df.append(series, ignore_index=True)
        return self

    def countries(self):
        """
        Return names of countries where records are registered.

        Raises:
            KeyError: Country names are not registered in this dataset

        Returns:
            list[str]: list of country names

        Notes:
            Country 'Others' will be removed.
        """
        country_list = super().countries()
        removed_countries = ["Others"]
        country_list = list(set(country_list) - set(removed_countries))
        return country_list


class Population(PopulationData):
    @deprecate(old="Population()", new="PopulationData()")
    def __init__(self, filename):
        super().__init__(filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import warnings
import pytest
import pandas as pd
from covsirphy import SubsetNotFoundError, PCRIncorrectPreconditionError
from covsirphy import COVID19DataHub, DataLoader, LinelistData
from covsirphy import Term, JHUData, CountryData, PopulationData
from covsirphy import OxCGRTData, PCRData, VaccineData


class TestCOVID19DataHub(object):
    def test_covid19dh(self):
        with pytest.raises(TypeError):
            COVID19DataHub(filename=None)
        data_hub = COVID19DataHub(
            filename=Path("input").joinpath("covid19dh.csv"))
        # Citation (with downloading), disabled to avoid downloading many times
        # assert isinstance(data_hub.primary, str)
        # Retrieve the dataset from the server
        data_hub.load(name="jhu", force=False, verbose=False)
        with pytest.raises(KeyError):
            data_hub.load(name="unknown")
        # Citation (without downloading)
        assert isinstance(data_hub.primary, str)


class TestDataLoader(object):
    def test_start(self):
        with pytest.raises(TypeError):
            DataLoader(directory=0)

    def test_dataloader(self, jhu_data, population_data, oxcgrt_data,
                        japan_data, linelist_data, pcr_data, vaccine_data):
        # List of primary sources of COVID-19 Data Hub
        data_loader = DataLoader()
        assert data_loader.covid19dh_citation
        # Data loading
        assert isinstance(jhu_data, JHUData)
        assert isinstance(population_data, PopulationData)
        assert isinstance(oxcgrt_data, OxCGRTData)
        assert isinstance(japan_data, CountryData)
        assert isinstance(linelist_data, LinelistData)
        assert isinstance(pcr_data, PCRData)
        assert isinstance(vaccine_data, VaccineData)
        # Local file
        data_loader.jhu(local_file="input/covid19dh.csv")
        data_loader.population(local_file="input/covid19dh.csv")
        data_loader.oxcgrt(local_file="input/covid19dh.csv")
        data_loader.pcr(local_file="input/covid19dh.csv")


class TestJHUData(object):
    def test_cleaning(self, jhu_data):
        assert isinstance(jhu_data.raw, pd.DataFrame)
        with pytest.raises(ValueError):
            jhu_data.cleaned(population=None)
        df = jhu_data.cleaned()
        assert set(df.columns) == set(Term.COLUMNS)
        assert isinstance(JHUData.from_dataframe(df), JHUData)

    def test_from_dataframe(self, japan_data):
        df = japan_data.cleaned()
        jhu_data_df = JHUData.from_dataframe(df)
        assert isinstance(jhu_data_df, JHUData)
        jhu_data_df.records("Japan")

    def test_subset(self, jhu_data):
        df = jhu_data.subset(
            "Japan", province="Tokyo", start_date="01Apr2020", end_date="01Jun2020")
        assert set(df.columns) == set(Term.NLOC_COLUMNS)
        with pytest.raises(KeyError):
            jhu_data.subset("Moon")
        with pytest.raises(ValueError):
            jhu_data.subset(
                "Japan", start_date="01Jan2020", end_date="10Jan2020")
        s_df = jhu_data.subset("Japan", population=126_500_000)
        assert set(s_df.columns) == set(Term.SUB_COLUMNS)
        jhu_data.subset("US")

    def test_replace(self, jhu_data, japan_data):
        jhu_data.replace(japan_data)
        df = jhu_data.subset("Japan")
        japan_df = japan_data.cleaned()
        last_date = japan_df.loc[japan_df.index[-1], Term.DATE]
        assert df.loc[df.index[-1], Term.DATE] == last_date

    def test_to_sr(self, jhu_data):
        df = jhu_data.to_sr("Japan", population=126_500_000)
        assert set(df.columns) == set([Term.R, Term.S])

    def test_total(self, jhu_data):
        df = jhu_data.total()
        column_set = set(Term.VALUE_COLUMNS) | set(Term.RATE_COLUMNS)
        assert set(df.columns) == column_set

    def test_countries(self, jhu_data):
        assert isinstance(jhu_data.countries(complement=False), list)
        assert isinstance(jhu_data.countries(complement=True), list)

    def test_closing_period(self, jhu_data):
        warnings.simplefilter("ignore", category=DeprecationWarning)
        assert isinstance(jhu_data.calculate_closing_period(), int)
        assert isinstance(jhu_data.calculate_recovery_period(), int)

    @pytest.mark.parametrize("country", ["UK"])
    def test_subset_complement_non_monotonic(self, jhu_data, country):
        df, is_complemented = jhu_data.subset_complement(country=country)
        assert is_complemented
        assert df[Term.C].is_monotonic_increasing

    @pytest.mark.parametrize("country", ["Netherlands", "China", "Germany"])
    def test_subset_complement_full(self, jhu_data, country):
        if country in set(["Netherlands", "China"]):
            with pytest.raises(ValueError):
                jhu_data.subset(country=country)
        df, is_complemented = jhu_data.subset_complement(country=country)
        assert isinstance(jhu_data.recovery_period, int)
        assert set(df.columns) == set(Term.NLOC_COLUMNS)
        assert is_complemented
        with pytest.raises(KeyError):
            jhu_data.subset_complement(country=country, end_date="01Jan1900")

    @pytest.mark.parametrize("country", ["Japan"])
    def test_subset_complement_partial(self, jhu_data, country):
        df, is_complemented = jhu_data.subset_complement(country=country)
        assert set(df.columns) == set(Term.NLOC_COLUMNS)
        assert is_complemented
        with pytest.raises(KeyError):
            jhu_data.subset_complement(country=country, end_date="01Jan1900")

    @pytest.mark.parametrize(
        "country", ["UK", "Netherlands", "China", "Germany", "France", "Japan"])
    def test_records(self, jhu_data, country):
        df, is_complemented = jhu_data.records(country=country)
        assert set(df.columns) == set(Term.NLOC_COLUMNS)
        assert is_complemented

    @pytest.mark.parametrize("country", ["Netherlands", "Moon"])
    def test_records_error(self, jhu_data, country):
        with pytest.raises(SubsetNotFoundError):
            jhu_data.records(country=country, auto_complement=False)

    @pytest.mark.parametrize(
        "applied, expected, iso3",
        [
            ("Congo", "Republic of the Congo", "COG"),
            ("Democratic Congo", "Democratic Republic of the Congo", "COD"),
            ("GR", "Greece", "GRC"),
            ("gr", "Greece", "GRC"),
            ("GRC", "Greece", "GRC"),
            ("Greece", "Greece", "GRC"),
            ("GREECE", "Greece", "GRC"),
            ("gre", "error", "GRC"),
            ("Ivory Coast", "Cote d'Ivoire", "CIV"),
            ("Korea, South", "South Korea", "KOR"),
            ("UK", "United Kingdom", "GBR"),
            ("US", "United States", "USA"),
            ("USA", "United States", "USA"),
            ("VAT", "Holy See", "VAT"),
        ]
    )
    def test_country_name(self, jhu_data, applied, expected, iso3):
        if expected == "error":
            with pytest.raises(KeyError):
                jhu_data.ensure_country_name(applied)
        else:
            response = jhu_data.ensure_country_name(applied)
            assert response == expected
            assert jhu_data.country_to_iso3(response) == iso3


class TestPopulationData(object):
    def test_cleaning(self, population_data):
        df = population_data.cleaned()
        column_set = set(Term.STR_COLUMNS) | set([Term.N, Term.ISO3])
        assert set(df.columns) == column_set

    def test_total(self, population_data):
        assert isinstance(population_data.total(), int)

    def test_to_dict(self, population_data):
        assert isinstance(population_data.to_dict(country_level=True), dict)
        assert isinstance(population_data.to_dict(country_level=False), dict)

    def test_value(self, population_data):
        assert isinstance(population_data.value("Japan"), int)
        old_value = population_data.value("Japan", date="01Mar2020")
        assert isinstance(old_value, int)
        with pytest.raises(KeyError):
            population_data.value("Japan", "01Jan1000")
        population_data.value("UK")

    def test_update(self):
        population_data = PopulationData(filename=None)
        population_data.update(1000, "Moon")
        assert population_data.value("Moon") == 1000
        population_data.update(2000, "Moon")
        assert population_data.value("Moon") == 2000

    def test_countries(self, population_data):
        assert isinstance(population_data.countries(), list)


class TestOxCGRTData(object):
    def test_cleaning(self, oxcgrt_data):
        df = oxcgrt_data.cleaned()
        assert set(df.columns) == set(OxCGRTData.OXCGRT_COLS)

    def test_subset(self, oxcgrt_data):
        with pytest.raises(KeyError):
            oxcgrt_data.subset("Moon")
        df = oxcgrt_data.subset("JPN")
        assert set(df.columns) == set(OxCGRTData.OXCGRT_COLS_WITHOUT_COUNTRY)

    def test_total(self, oxcgrt_data):
        with pytest.raises(NotImplementedError):
            oxcgrt_data.total()


class TestPCRData(object):
    def test_cleaning(self, pcr_data):
        df = pcr_data.cleaned()
        assert set(df.columns) == set(PCRData.PCR_COLUMNS)

    def test_from_dataframe(self, pcr_data):
        df = pcr_data.cleaned()
        assert isinstance(PCRData.from_dataframe(df), PCRData)

    def test_use_ourworldindata(self, pcr_data):
        pcr_data.use_ourworldindata(
            filename="input/ourworldindata_pcr.csv")

    def test_subset(self, pcr_data):
        with pytest.raises(SubsetNotFoundError):
            pcr_data.subset("Greece", end_date="01Jan2000")
        df = pcr_data.subset("Greece")
        assert set(df.columns) == set(PCRData.PCR_NLOC_COLUMNS)

    def test_subset_complement(self, pcr_data):
        with pytest.raises(NotImplementedError):
            pcr_data.subset_complement("Greece")

    def test_records(self, pcr_data):
        with pytest.raises(SubsetNotFoundError):
            pcr_data.records("Greece", end_date="01Jan2000")
        df, _ = pcr_data.records("Greece")
        assert set(df.columns) == set(PCRData.PCR_NLOC_COLUMNS)

    @pytest.mark.parametrize("country", ["Greece", "Italy", "Sweden"])
    def test_positive_rate(self, pcr_data, country):
        warnings.simplefilter("ignore", category=UserWarning)
        pcr_data.positive_rate(country, show_figure=True)
        df = pcr_data.positive_rate(country, show_figure=False)
        assert set(
            [PCRData.T_DIFF, PCRData.C_DIFF, PCRData.PCR_RATE]).issubset(df.columns)

    @pytest.mark.parametrize("country", ["China"])
    def test_positive_rate_error(self, pcr_data, country):
        with pytest.raises(PCRIncorrectPreconditionError):
            pcr_data.positive_rate(country, show_figure=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
import pytest
from covsirphy import ExampleData, PopulationData, Term, ModelValidator
from covsirphy import ModelBase, SIR, SIRD, SIRF, SIRFV, SEWIRF


class TestODE(object):
    @pytest.mark.parametrize(
        "model",
        [SIR, SIRD, SIRF, SIRFV, SEWIRF])
    def test_ode(self, model):
        # Setting
        eg_tau = 1440
        area = {"country": "Full", "province": model.NAME}
        # Population
        population_data = PopulationData(filename=None)
        population_data.update(model.EXAMPLE["population"], **area)
        # Simulation
        example_data = ExampleData(tau=eg_tau, start_date="01Jan2020")
        example_data.add(model, **area)
        # Model-specialized records
        spe_df = example_data.specialized(**area)
        assert set(spe_df.columns) == set(
            [*Term.STR_COLUMNS, *model.VARIABLES])
        # Non-dimensional records
        nondim_df = example_data.non_dim(**area)
        assert set(nondim_df.columns) == set(
            [Term.TS, *list(model.VAR_DICT.keys())])
        # JHU-type records
        jhu_df = example_data.subset(**area)
        assert set(jhu_df.columns) == set(Term.NLOC_COLUMNS)
        # Calculate Rt/day parameters when parameters are None
        param_dict = {p: 0 for p in model.PARAMETERS}
        model_instance = model(population_data.value(**area), **param_dict)
        model_instance.calc_r0()
        model_instance.calc_days_dict(eg_tau)

    @pytest.mark.parametrize("model", [SIR])
    def test_model_common(self, model):
        model_ins = model(population=1_000_000, rho=0.2, sigma=0.075)
        assert str(model_ins) == "SIR model with rho=0.2, sigma=0.075"
        assert model_ins["rho"] == 0.2
        with pytest.raises(KeyError):
            assert model_ins["kappa"] == 0.1

    @pytest.mark.parametrize("model", [ModelBase])
    def test_model_base(self, model):
        model_ins = model(population=1_000_000)
        with pytest.raises(NotImplementedError):
            model_ins(1, [0, 0, 0])
        with pytest.raises(NotImplementedError):
            model.param_range(1, 2)
        with pytest.raises(NotImplementedError):
            model.specialize(1, 2)
        with pytest.raises(NotImplementedError):
            model_ins.calc_r0()
        with pytest.raises(NotImplementedError):
            model_ins.calc_days_dict(1440)

    @pytest.mark.parametrize("model", [SIR])
    def test_error(self, model):
        # Setting
        eg_tau = 1440
        # Simulation
        example_data = ExampleData(tau=eg_tau, start_date="01Jan2020")
        with pytest.raises(KeyError):
            assert not example_data.specialized(model=model).empty
        with pytest.raises(KeyError):
            assert not example_data.non_dim(model=model).empty
        example_data.add(model)
        # Model-specialized records
        with pytest.raises(ValueError):
            assert not example_data.specialized().empty

    def test_ode_two_phases(self, population_data):
        # Setting
        eg_tau = 1440
        area = {"country": "Example", "province": "Example"}
        # Simulation
        example_data = ExampleData(tau=eg_tau)
        example_data.add(SIRF, step_n=30, **area)
        example_data.add(SIRD, step_n=30, **area)
        dim_df = example_data.subset(**area)
        assert isinstance(dim_df, pd.DataFrame)
        assert set(dim_df.columns) == set(Term.NLOC_COLUMNS)

    @pytest.mark.parametrize("model", [SIR])
    def test_validation_sir(self, model):
        # Setting
        validator = ModelValidator(n_trials=4, seed=1)
        # Execute validation
        validator.run(model, timeout=10)
        validator.summary()
        with pytest.raises(ValueError):
            validator.run(model)

    @pytest.mark.parametrize("model", [SIRD, SIRF])
    def test_validation(self, model):
        # Setting
        validator = ModelValidator(n_trials=1, seed=1)
        # Execute validation
        validator.run(model, timeout=10)
        validator.summary()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import numpy as np
import pandas as pd
from covsirphy.util.error import SubsetNotFoundError
from covsirphy.cleaning.cbase import CleaningBase


class VaccineData(CleaningBase):
    """
    Dataset regarding vaccination retrieved from "Our World In Data".
    https://github.com/owid/covid-19-data/tree/master/public/data
    https://ourworldindata.org/coronavirus

    Args:
        filename (str or pathlib.path): CSV filename to save the raw dataset
        force (bool): if True, always download the dataset from the server
        verbose (int): level of verbosity
    Notes:
        Columns of VaccineData.cleaned():
            - Date (pandas.TimeStamp): observation dates
            - Country (pandas.Category): country (or province) names
            - Product (pandas.Category): product names
            - Vaccinations (int): the number of vaccinations
    """
    # URL
    URL = "https://covid.ourworldindata.org/data/vaccinations/"
    URL_REC = f"{URL}vaccinations.csv"
    URL_LOC = f"{URL}locations.csv"
    # Columns
    VAC_COLS = [
        CleaningBase.DATE, CleaningBase.COUNTRY, CleaningBase.PRODUCT, CleaningBase.VAC]

    def __init__(self, filename, force=False, verbose=1):
        Path(filename).parent.mkdir(exist_ok=True, parents=True)
        if Path(filename).exists() and not force:
            self._raw = self.load(filename)
        else:
            self._raw = self._retrieve(filename=filename, verbose=verbose)
        self._cleaned_df = self._cleaning()
        self._citation = "Hasell, J., Mathieu, E., Beltekian, D. et al." \
            " A cross-country database of COVID-19 testing. Sci Data 7, 345 (2020)." \
            " https://doi.org/10.1038/s41597-020-00688-8"

    def _retrieve(self, filename, verbose=1):
        """
        Retrieve the dataset from server.
        Args:
            filename (str or pathlib.path): CSV filename to save the raw dataset
            verbose (int): level of verbosity
        Returns:
            pd.DataFrame:
                Index: reset index
                Columns: Date, Country, Product, Vaccinations
        """
        # Show URL
        if verbose:
            print(
                "Retrieving COVID-19 vaccination dataset from https://covid.ourworldindata.org/data/")
        # Download datasets and merge them
        rec_df = self.load(
            self.URL_REC, columns=["location", "date", "total_vaccinations"])
        loc_df = self.load(self.URL_LOC, columns=["location", "vaccines"])
        df = rec_df.merge(loc_df, how="left", on="location")
        df = df.rename(
            {
                "vaccines": self.PRODUCT, "total_vaccinations": self.VAC,
                "date": self.DATE, "location": self.COUNTRY}, axis=1)
        # Remove "World" records (total values)
        df = df.loc[df[self.COUNTRY] != "World"]
        # Save the dataframe as CSV file
        df.to_csv(filename, index=False)
        return df

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.

        Returns:
            pandas.DataFrame:
                Index: reset index
                Columns: Date, Country, Product, Vaccinations
        """
        df = self._raw.copy()
        # Date
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        # Location (Country)
        df[self.COUNTRY] = df[self.COUNTRY].astype("category")
        # Company
        df[self.PRODUCT] = df[self.PRODUCT].astype("category")
        # Vaccinations
        df[self.VAC] = pd.to_numeric(df[self.VAC], errors="coerce")
        df[self.VAC] = df[self.VAC].fillna(method="ffill").fillna(0)
        df[self.VAC] = df[self.VAC].astype(np.int64)
        return df.loc[:, self.VAC_COLS]

    def subset(self, country, product=None, start_date=None, end_date=None):
        """
        Return subset of the country/province and start/end date.

        Args:
            country (str or None): country name or ISO3 code
            product (str or None): product name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020

        Returns:
            pandas.DataFrame
                Index: reset index
                Columns:
                    - Date (pandas.TimeStamp): observation date
                    - Vaccinations (int): the number of vaccinations
        """
        df = self._cleaned_df.copy()
        # Subset by country
        country_alias = self.ensure_country_name(country)
        df = df.loc[df[self.COUNTRY] == country_alias]
        # Subset by product name
        if product is not None:
            df = df.loc[df[self.PRODUCT] == product]
        # Subset with start date
        if start_date is not None:
            df = df.loc[df[self.DATE] >= self.date_obj(start_date)]
        # Subset with end date
        if end_date is not None:
            df = df.loc[df[self.DATE] <= self.date_obj(end_date)]
        # Resampling
        df = df.set_index(self.DATE).resample("D").sum().reset_index()
        # Fill in the blanks
        df[self.VAC] = df[self.VAC].replace(0, None)
        df[self.VAC] = df[self.VAC].fillna(method="ffill").fillna(0)
        # Check records were found
        if df.empty:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=product,
                start_date=start_date, end_date=end_date)
        return df

    def records(self, country, product=None, start_date=None, end_date=None):
        """
        Return subset of the country/province and start/end date.

        Args:
            country (str or None): country name or ISO3 code
            product (str or None): product name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020

        Returns:
            pandas.DataFrame
                Index: reset index
                Columns:
                    - Date (pandas.TimeStamp): observation date
                    - Vaccinations (int): the number of vaccinations
        """
        return self.subset(
            country=country, product=product, start_date=start_date, end_date=end_date)

    def total(self):
        """
        Calculate total values of the cleaned dataset.

        Returns:
            pandas.DataFrame
                Index: reset index
                Columns:
                    - Date (pandas.TimeStamp): observation date
                    - Vaccinations (int): the number of vaccinations
        """
        df = self._cleaned_df.copy()
        # Remove duplication
        df = df.loc[df[self.COUNTRY] == "United Kingdom"]
        # Resampling
        df = df.set_index(self.DATE).resample("D").sum()
        return df.reset_index()
